/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "Database.h"
#include "version.h"
#include "DefaultValues.h"

// Qt includes
#include <QDebug>
#include <QSqlQuery>
#include <QMessageBox>
#include <QDir>
#include <QSqlError>

namespace {

// The database version we're using
constexpr int DB_VERSION = 5;

// Database string literals
const QString DB_DB_VERSION          = QString::fromUtf8("DB_VERSION");
const QString DB_MUCKTURNIER_VERSION = QString::fromUtf8("MUCKTURNIER_VERSION");
const QString DB_TOURNAMENT_MODE     = QString::fromUtf8("TOURNAMENT_MODE");
const QString DB_BOOGER_SCORE        = QString::fromUtf8("BOOGER_SCORE");
const QString DB_BOOGERS_PER_ROUND   = QString::fromUtf8("BOOGERS_PER_ROUND");
const QString DB_FIXED_PAIRS         = QString::fromUtf8("FIXED_PAIRS");
const QString DB_SINGLE_PLAYERS      = QString::fromUtf8("SINGLE_PLAYERS");

const QString DB_CONNECTION_NAME = QString::fromUtf8("SQLiteDB");

}

Database::Database(QObject *parent) : QObject(parent)
{
}

Database::~Database()
{
    closeDatabaseConnection();
}

void Database::closeDatabaseConnection()
{
    if (m_db.isOpen()) {
        m_db.commit();
        m_db.close();
        m_db = QSqlDatabase::database();
        QSqlDatabase::removeDatabase(DB_CONNECTION_NAME);
    }
}

bool Database::connectDatabase(QString &dbFile)
{
    closeDatabaseConnection();

    m_db = QSqlDatabase::addDatabase(QString::fromUtf8("QSQLITE"), DB_CONNECTION_NAME);
    m_db.setDatabaseName(dbFile);

    if (! m_db.open()) {
        // This should not happen
        QMessageBox::critical(dynamic_cast<QWidget *>(parent()),
            tr("Datenbankverbindung fehlgeschlagen"),
            tr("Unerwarteter Fehler: Konnte keine Verbindung zur Datenbank „%1“ herstellen.").arg(
               QDir::toNativeSeparators(dbFile)),
            QMessageBox::Ok);
        return false;
    }

    return true;
}

bool Database::create(QString dbFile)
{
    if (! connectDatabase(dbFile)) {
        return false;
    }

    m_db.transaction();
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("CREATE TABLE config(\n"
                                 "    key TEXT,\n"
                                 "    value TEXT\n"
                                 ")"));
    query.exec(QString::fromUtf8("CREATE TABLE players(\n"
                                 "    id INTEGER PRIMARY KEY,\n"
                                 "    name TEXT\n"
                                 ")"));
    query.exec(QString::fromUtf8("CREATE TABLE assignment(\n"
                                 "    round INTEGER,\n"
                                 "    pair_or_player INTEGER,\n"
                                 "    pair_number INTEGER,\n"
                                 "    player_number INTEGER,\n"
                                 "    table_number INTEGER\n"
                                 ")"));
    query.exec(QString::fromUtf8("CREATE TABLE score(\n"
                                 "    round INTEGER,\n"
                                 "    table_number INTEGER,\n"
                                 "    booger INTEGER,\n"
                                 "    pair_or_player INTEGER,\n"
                                 "    score INTEGER\n"
                                 ")"));

    query.prepare(QString::fromUtf8("INSERT INTO config(key, value) VALUES(?, ?)"));

    QVariantList keys;
    keys << DB_MUCKTURNIER_VERSION
         << DB_DB_VERSION
         << DB_TOURNAMENT_MODE
         << DB_BOOGER_SCORE
         << DB_BOOGERS_PER_ROUND;
    query.addBindValue(keys);

    QVariantList values;
    values << QString::fromUtf8(MT_VERSION)
           << DB_VERSION
           << DB_FIXED_PAIRS
           << DEFAULT_BOOGER_SCORE
           << DEFAULT_BOOGERS_PER_ROUND;
    query.addBindValue(values);
    query.execBatch();

    m_db.commit();

    updateSettings();
    m_tournamentStarted = false;
    return true;
}

void Database::updateSettings()
{
    // Reads the settings stored in the database.

    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT key, value FROM config"));

    while (query.next()) {
        QString key = query.value(0).toString();
        QString value = query.value(1).toString();

        if (key == DB_MUCKTURNIER_VERSION) {
            m_dbValueMuckturnierVersion = value;
        } else if (key == DB_TOURNAMENT_MODE) {
            if (value == DB_FIXED_PAIRS) {
                m_settingsTournamentMode = TournamentMode::fixedPairs;
            } else if (value == DB_SINGLE_PLAYERS) {
                m_settingsTournamentMode = TournamentMode::singlePlayers;
            }
        } else if (key == DB_BOOGER_SCORE) {
            m_settingsBoogerScore = value.toInt();
        } else if (key == DB_BOOGERS_PER_ROUND) {
            m_settingsBoogersPerRound = value.toInt();
        }
    }
}

bool Database::tournamentStarted() const
{
    return m_tournamentStarted;
}

bool Database::openDb(QString dbFile)
{
    m_openError.clear();
    m_openErrorCode = OpenError::noError;

    if (! connectDatabase(dbFile)) {
        return false;
    }

    QSqlQuery query(m_db);

    // Let's see if we have all tables
    query.exec(QString::fromUtf8("SELECT name "
                                 "FROM sqlite_master "
                                 "WHERE "
                                     "type = 'table' "
                                     "AND name = 'config' "
                                     "OR name = 'players' "
                                     "OR name = 'assignment' "
                                     "OR name = 'score'"));
    int tableCount = 0;
    while (query.next()) {
        tableCount++;
    }
    if (tableCount != 4) {
        m_openError = tr("In der Datenbank sind nicht alle nötigen Tabllen vorhanden. "
                         "vermutlich ist „%1“ keine Muckturnier-Datenbank, oder die Datei wurde "
                         "mit einer zu alten Version von Muckturnier (vor 0.5.1) erstellt!").arg(
                         QDir::toNativeSeparators(dbFile));
        m_openErrorCode = OpenError::tablesMissing;
        return false;
    }

    // Let's see if we have the correct database version
    query.prepare(QString::fromUtf8("SELECT value FROM config WHERE key = ?"));
    query.bindValue(0, DB_DB_VERSION);
    query.exec();
    query.next();
    int dbVersion = query.value(0).toInt();
    if (dbVersion != DB_VERSION) {
        if (dbVersion < DB_VERSION) {
            m_openError = tr("<p>Die Datenbank hat die falsche Version (ist %1, sollte %2 sein)."
                             "</p>"
                             "<p>Entweder wurde das Turnier mit einer zu alten Version von "
                             "Muckturnier (vor 0.5.1) erstellt, oder die Datei ist keine "
                             "Muckturnier-Datenbank.</p>").arg(
                             QString::number(dbVersion), QString::number(DB_VERSION));
        } else {
            m_openError = tr("<p>Die Datenbank hat die falsche Version (ist %1, sollte %2 sein). "
                             "Das Tunier wurde vermutlich mit einer neueren Version von "
                             "Muckturnier erstellt.</p>"
                             "<p>Bitte auf die neueste Version aktualisieren (zu finden unter "
                             "<a href='http://nasauber.de/opensource/muckturnier/'>"
                             "http://nasauber.de/opensource/muckturnier/</a>)!</p>").arg(
                             QString::number(dbVersion), QString::number(DB_VERSION));
        }
        if (dbVersion == 3) {
            m_openErrorCode = OpenError::version3Db;
        } else if (dbVersion == 4) {
            m_openErrorCode = OpenError::version4Db;
        } else {
            m_openErrorCode = OpenError::wrongVersion;
        }
        return false;
    }

    // Everything seems to be fine :-)
    updateSettings();
    checkIfTournamentStarted();

    return true;
}

QString Database::openError() const
{
    return m_openError;
}

Database::OpenError Database::openErrorCode() const
{
    return m_openErrorCode;
}

QString Database::playersString(Numerus numerus, bool article) const
{
    switch (m_settingsTournamentMode) {
    case TournamentMode::fixedPairs:
        switch (numerus) {
        case Numerus::singular:
            return article ? tr("das Paar")  : tr("Paar");
        case Numerus::plural:
            return article ? tr("die Paare") : tr("Paare");
        }
    case TournamentMode::singlePlayers:
        switch (numerus) {
        case Numerus::singular:
            return article ? tr("der Spieler") : tr("Spieler");
        case Numerus::plural:
            return article ? tr("die Spieler") : tr("Spieler");
        }
    }
    // This is just to make the compiler happy. We can't reach here.
    return QString();
}

void Database::reset()
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("DELETE FROM players"));
    query.exec(QString::fromUtf8("DELETE FROM assignment"));
    query.exec(QString::fromUtf8("DELETE FROM score"));
    m_db.commit();
    m_tournamentStarted = false;
}

void Database::saveSettings()
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("UPDATE config SET value = ? WHERE key = ?"));

    QVariantList values;
    switch (m_settingsTournamentMode) {
    case TournamentMode::fixedPairs:
        values << DB_FIXED_PAIRS;
        break;
    case TournamentMode::singlePlayers:
        values << DB_SINGLE_PLAYERS;
        break;
    }
    values << m_settingsBoogerScore
           << m_settingsBoogersPerRound;
    query.addBindValue(values);

    QVariantList keys;
    keys << DB_TOURNAMENT_MODE
         << DB_BOOGER_SCORE
         << DB_BOOGERS_PER_ROUND;
    query.addBindValue(keys);

    query.execBatch();
    m_db.commit();
}

void Database::checkIfTournamentStarted()
{
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT COUNT(round) FROM SCORE"));
    query.next();
    m_tournamentStarted = query.value(0).toInt() != 0;
}

int Database::necessaryDivisor() const
{
    return 2 + (m_settingsTournamentMode != TournamentMode::fixedPairs) * 2;
}

QStringList Database::listPlayers()
{
    // Returns an alphabetically sorted list of all player/pair names.

    QStringList playersList;
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT name FROM players ORDER BY name COLLATE NOCASE"));
    while (query.next()) {
        playersList << query.value(0).toString();
    }

    return playersList;
}

bool Database::checkName(QString name)
{
    // Checks if a player/pair name already exists.
    name = name.toLower();
    for (const QString &player : listPlayers()) {
        if (player.toLower() == name) {
            return false;
        }
    }
    return true;
}

void Database::registerPlayers(QString name)
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("INSERT INTO players(id, name) VALUES(NULL, ?)"));
    query.bindValue(0, name);
    query.exec();
    m_db.commit();
}

void Database::registerPlayers(QStringList &names)
{
    QVariantList allNames;
    for (const QString &name : names) {
        allNames << name;
    }

    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("INSERT INTO players(id, name) VALUES(NULL, ?)"));
    query.addBindValue(allNames);
    query.execBatch();
    m_db.commit();
}

void Database::deletePlayers(QString name)
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("DELETE FROM players WHERE name = ?"));
    query.bindValue(0, name);
    query.exec();
    m_db.commit();
}

void Database::renamePlayers(QString oldName, QString newName)
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("UPDATE players SET name = ? WHERE name = ?"));
    query.bindValue(0, newName);
    query.bindValue(1, oldName);
    query.exec();
    m_db.commit();
}

int Database::playersCount()
{
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT COUNT(id) AS count FROM players"));
    query.next();
    return query.value(0).toInt();
}


QMap<int, bool> Database::roundsState()
{
    QMap<int, bool> roundsStateMap;
    int allBoogers = playersCount() * m_settingsBoogersPerRound;
    int roundNumber = 1;

    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT round, COUNT(round) AS count FROM score "
                                 "GROUP BY round ORDER BY round"));

    // Check if each round already contains all boogers
    while (query.next()) {
        roundNumber = query.value(0).toInt();
        if (query.value(1).toInt() < allBoogers) {
            roundsStateMap[roundNumber] = false;
        } else {
            roundsStateMap[roundNumber] = true;
        }
    }

    // Be sure to have a state for each round. All data from a round could have been deleted.
    // If we have no data, round 1 will be added as open.
    for (int i = 1; i <= roundNumber; i++) {
        if (! roundsStateMap.contains(i)) {
            roundsStateMap[i] = false;
        }
    }

    return roundsStateMap;
}

QList<int> Database::missingTables(int round)
{
    QList<int> missingTableNumbers;

    // First, we assume we need all tables
    for (int i = 1; i <= playersCount() / necessaryDivisor(); i++) {
        missingTableNumbers << i;
    }

    // Then, we take away those that already have a score
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("SELECT table_number FROM score WHERE round = ? "
                                    "GROUP BY table_number ORDER BY table_number"));
    query.bindValue(0, round);
    query.exec();
    while (query.next()) {
        missingTableNumbers.removeOne(query.value(0).toInt());
    }

    return missingTableNumbers;
}

QMap<QString, int> Database::missingPlayers(int round)
{
    QMap<QString, int> missingPlayersMap;
    QSqlQuery query(m_db);

    // Get a list of all players/pairs
    query.exec(QString::fromUtf8("SELECT name, id FROM players"));
    while (query.next()) {
        missingPlayersMap[query.value(0).toString()] = query.value(1).toInt();
    }

    // Get a list of all players/pairs with a score in this round and remove them
    query.prepare(QString::fromUtf8("SELECT pair_or_player FROM score "
                                    "WHERE round = ? GROUP BY pair_or_player"));
    query.bindValue(0, round);
    query.exec();
    while (query.next()) {
        missingPlayersMap.remove(missingPlayersMap.key(query.value(0).toInt()));
    }

    return missingPlayersMap;
}

Database::Assignment Database::getAssignment(int roundNumber)
{
    int round = roundNumber - 1;
    int allTables = playersCount() / necessaryDivisor();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("SELECT pair_or_player, pair_number, table_number "
                                    "FROM assignment WHERE round = ?"));
    query.bindValue(0, round);
    query.exec();

    QMap<int, int> assignedTable;
    QMap<int, int> assignedPairNumber;
    QMap<int, QMap<int, int>> assignedPair;

    while(query.next()) {
        if (query.value(1).toInt() == 1) {
            assignedTable[query.value(0).toInt()] = query.value(2).toInt();
        } else {
            assignedTable[query.value(0).toInt()] = query.value(2).toInt() % allTables + 1;
        }
        assignedPairNumber[query.value(0).toInt()] = query.value(1).toInt();
    }

    for (const int &pairNumber : assignedTable.keys()) {
        assignedPair[assignedTable[pairNumber]][assignedPairNumber[pairNumber]] = pairNumber;
    }

    Assignment assignment;
    assignment.assignedTable = assignedTable;
    assignment.assignedPairNumber = assignedPairNumber;
    assignment.assignedPair = assignedPair;

    return assignment;
}

QList<Database::Score> Database::getScores(int round)
{
    QSqlQuery query(m_db);
    switch (m_settingsTournamentMode) {
    case TournamentMode::fixedPairs:
        query.prepare(QString::fromUtf8(
            "SELECT "
                "score.table_number, "
                "players.name, "
                "score.score "
            "FROM score "
            "INNER JOIN players ON "
                "players.id = score.pair_or_player "
            "INNER JOIN assignment ON "
                "assignment.round = score.round "
                "AND assignment.pair_or_player = score.pair_or_player "
            "WHERE score.round = ? "
            "ORDER BY "
                "score.table_number, "
                "assignment.pair_number, "
                "score.booger"
        ));
        break;
    case TournamentMode::singlePlayers:
        // It's a bit more complicated for single players ;-)
        query.prepare(QString::fromUtf8(
            "SELECT "
                "score_one.table_number, "
                "player_one.name || ' / ' || player_two.name, "
                "score_one.score "
            "FROM "
                "score AS score_one "
            "INNER JOIN players AS player_one ON "
                "player_one.id = score_one.pair_or_player "
            "INNER JOIN assignment AS assignment_one ON "
                "assignment_one.round = score_one.round "
                "AND assignment_one.table_number = score_one.table_number "
                "AND assignment_one.pair_or_player = score_one.pair_or_player "
            "INNER JOIN players AS player_two ON "
                "player_two.id = score_two.pair_or_player "
            "INNER JOIN score AS score_two ON "
                "score_two.round = score_one.round "
                "AND score_two.table_number = score_one.table_number "
                "AND score_two.booger = score_one.booger "
                "AND assignment_two.pair_number = assignment_one.pair_number "
                "AND assignment_two.player_number = 2 "
            "INNER JOIN assignment AS assignment_two ON "
                "assignment_two.round = score_one.round "
                "AND assignment_two.table_number = score_one.table_number "
                "AND assignment_two.pair_or_player = score_two.pair_or_player "
            "WHERE "
                "score_one.round = ? "
                "AND assignment_one.player_number = 1 "
            "ORDER BY "
                "score_one.table_number, "
                "assignment_one.pair_number, "
                "score_one.booger "
        ));
        break;
    }
    query.bindValue(0, round);
    query.exec();

    QList<Score> scores;
    while (query.next()) {
        Score score;
        score.table = query.value(0).toInt();
        score.pairName = query.value(1).toString();
        score.score = query.value(2).toInt();
        scores << score;
    }

    return scores;
}

QList<Database::SingleScore> Database::getSingleScore(int round, int table)
{
    QList<SingleScore> score;
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8(
        "SELECT "
            "score.pair_or_player, "
            "score.score "
        "FROM score "
        "INNER JOIN assignment ON "
            "assignment.round = score.round "
            "AND assignment.pair_or_player = score.pair_or_player "
        "WHERE "
            "score.round = ? "
            "AND score.table_number = ? "
        "ORDER BY "
            "assignment.pair_number, "
            "assignment.player_number, "
            "score.booger"
    ));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.exec();

    while (query.next()) {
        SingleScore singleScore;
        singleScore.pairOrPlayer = query.value(0).toInt();
        singleScore.score = query.value(1).toInt();
        score << singleScore;
    }

    return score;
}

void Database::insertScore(QSqlQuery &query,
                           int round, int table, int booger, int pairOrPlayer, int score)
{
    query.prepare(QString::fromUtf8("INSERT INTO score(round, table_number, booger, "
                                                      "pair_or_player, score) "
                                    "VALUES(?, ?, ?, ?, ?)"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.bindValue(2, booger);
    query.bindValue(3, pairOrPlayer);
    query.bindValue(4, score);
    query.exec();
}

void Database::updateAssignment(QSqlQuery &query, int round, int pairOrPlayer,
                                                  int pairNumber, int playerNumber, int table)
{
    query.prepare(QString::fromUtf8(
        "INSERT INTO assignment(round, pair_or_player, pair_number, player_number, table_number) "
        "VALUES(?, ?, ?, ?, ?)"
    ));
    query.bindValue(0, round);
    query.bindValue(1, pairOrPlayer);
    query.bindValue(2, pairNumber);
    query.bindValue(3, playerNumber);
    query.bindValue(4, table);
    query.exec();
}

void Database::saveScore(int round, int table,
                         QList<int> players, QList<int> pair1Scores, QList<int> pair2Scores)
{
    m_db.transaction();
    QSqlQuery query(m_db);

    for (int booger = 1; booger <= pair1Scores.size(); booger++) {
        insertScore(query, round, table, booger, players.at(0), pair1Scores.at(booger - 1));
        insertScore(query, round, table, booger, players.at(1), pair2Scores.at(booger - 1));
        if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
            insertScore(query, round, table, booger, players.at(2), pair1Scores.at(booger - 1));
            insertScore(query, round, table, booger, players.at(3), pair2Scores.at(booger - 1));
        }

        if (booger == 1) {
            updateAssignment(query, round, players.at(0), 1, 1, table);
            updateAssignment(query, round, players.at(1), 2, 1, table);
            if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
                updateAssignment(query, round, players.at(2), 1, 2, table);
                updateAssignment(query, round, players.at(3), 2, 2, table);
            }
        }
    }

    m_db.commit();

    if (! m_tournamentStarted) {
        m_tournamentStarted = true;
    }
}

void Database::deleteScore(int round, int table)
{
    m_db.transaction();
    QSqlQuery query(m_db);

    query.prepare(QString::fromUtf8("DELETE FROM score WHERE round = ? AND table_number = ?"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.exec();

    query.prepare(QString::fromUtf8("DELETE FROM assignment WHERE round = ? AND table_number = ?"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.exec();

    m_db.commit();
    checkIfTournamentStarted();
}

QList<Database::Ranking> Database::getRanking(int upToRound)
{
    QSqlQuery query(m_db);
    QString rankingQuery = QString::fromUtf8(
        "SELECT "
            "players.name, "
            "SUM(CASE WHEN one_score.score = ? THEN 1 ELSE 0 END) AS won_boogers, "
            "SUM(CASE WHEN one_score.score = ? THEN 0 ELSE one_score.score END) AS scored_goals, "
            "SUM(CASE WHEN other_score.score = ? THEN 0 ELSE other_score.score END) AS opponent_goals, "
            "COUNT(one_score.score) AS played_boogers "
        "FROM "
            "score AS one_score "
        "INNER JOIN players ON "
            "players.id = one_score.pair_or_player "
        "INNER JOIN assignment AS one_assignment ON "
            "one_assignment.round = one_score.round "
            "AND one_assignment.table_number = one_score.table_number "
            "AND one_assignment.pair_or_player = one_score.pair_or_player "
        "INNER JOIN assignment AS other_assignment ON "
            "other_assignment.round = other_score.round "
            "AND other_assignment.table_number = other_score.table_number "
            "AND other_assignment.pair_or_player = other_score.pair_or_player "
        "INNER JOIN score AS other_score ON "
            "other_score.round = one_score.round "
            "AND other_score.table_number = one_score.table_number "
            "AND other_score.booger = one_score.booger "
            "AND other_assignment.pair_number != one_assignment.pair_number "
        "WHERE one_score.round <= ? "
        "GROUP BY "
            "one_score.pair_or_player "
        "ORDER BY "
            "won_boogers DESC, "
            "scored_goals DESC, "
            "opponent_goals, "
            "players.name"
    );
    query.prepare(rankingQuery);
    for (int i = 0; i < 3; i++) {
        query.bindValue(i, m_settingsBoogerScore);
    }
    query.bindValue(3, upToRound);
    query.exec();

    QList<Ranking> allRankings;
    QStringList namesWithScore;
    while (query.next()) {
        Ranking ranking;
        ranking.name = query.value(0).toString();
        ranking.boogers = query.value(1).toInt();
        ranking.goals = query.value(2).toInt();
        ranking.opponentGoals = query.value(3).toInt();
        ranking.playedBoogers = query.value(4).toInt();
        if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
            ranking.boogers /= 2;
            ranking.goals /= 2;
            ranking.opponentGoals /= 2;
            ranking.playedBoogers /= 2;
        }
        allRankings << ranking;
        namesWithScore << ranking.name;
    }

    // Be sure to have data for all players/pairs, in case some of them don't have a score yet
    for (const QString &name : listPlayers()) {
        if (! namesWithScore.contains(name)) {
            Ranking ranking;
            ranking.name = name;
            ranking.boogers = 0;
            ranking.goals = 0;
            ranking.opponentGoals = 0;
            ranking.playedBoogers = 0;
            allRankings << ranking;
        }
    }

    return allRankings;
}

QString Database::getMuckturnierVersion() const
{
    return m_dbValueMuckturnierVersion;
}

Database::TournamentMode Database::getTournamentMode() const
{
    return m_settingsTournamentMode;
}

QString Database::getDbFile() const
{
    return m_db.databaseName();
}

void Database::setTournamentMode(Database::TournamentMode tournamentMode)
{
    m_settingsTournamentMode = tournamentMode;
}

int Database::getDbVersion() const
{
    return DB_VERSION;
}

int Database::getBoogerScore() const
{
    return m_settingsBoogerScore;
}

void Database::setBoogerScore(int boogerScore)
{
    m_settingsBoogerScore = boogerScore;
}

int Database::getBoogersPerRound() const
{
    return m_settingsBoogersPerRound;
}

void Database::setBoogersPerRound(int boogersPerRound)
{
    m_settingsBoogersPerRound = boogersPerRound;
}

QString Database::boogify(int score)
{
    if (score == m_settingsBoogerScore) {
        return QString::number(score);
    } else {
        return QString::fromUtf8("%1 ●").arg(QString::number(score));
    }
}

void Database::updateDatabase3to4(QString dbFileName)
{
    connectDatabase(dbFileName);
    QSqlQuery query(m_db);
    m_db.transaction();

    // Get the tournament mode
    query.exec(QString::fromUtf8("SELECT value FROM config WHERE key = 'TOURNAMENT_MODE'"));
    query.next();
    QString tournamentMode = query.value(0).toString();

    // Add the missing column to "assignment"
    query.exec(QString::fromUtf8("ALTER TABLE assignment ADD COLUMN player_number INT"));

    // Fill in the data. For a single player mode database, the fetch result hopefully is still in
    // the correct order, so that the right players become player 1 and 2 inside a team.
    if (tournamentMode == QString::fromUtf8("FIXED_PAIRS")) {
        query.exec(QString::fromUtf8("UPDATE assignment SET player_number = 1"));
    } else {
        QVariantList round;
        QVariantList pair_or_player;
        QVariantList pair_number;
        QVariantList table_number;
        QVariantList player_number;
        query.exec(QString::fromUtf8("SELECT round, pair_or_player, pair_number, table_number "
                                     "FROM assignment "
                                     "ORDER BY round, table_number, pair_number"));
        int playerNumber = 1;
        while (query.next()) {
            round << query.value(0).toInt();
            pair_or_player << query.value(1).toInt();
            pair_number << query.value(2).toInt();
            table_number << query.value(3).toInt();
            playerNumber = (playerNumber + 1) % 2;
            player_number << playerNumber + 1;
        }

        query.prepare(QString::fromUtf8("UPDATE assignment SET player_number = ? WHERE "
                                        "round = ? "
                                        "AND pair_or_player = ? "
                                        "AND pair_number = ? "
                                        "AND table_number = ?"));
        query.addBindValue(player_number);
        query.addBindValue(round);
        query.addBindValue(pair_or_player);
        query.addBindValue(pair_number);
        query.addBindValue(table_number);
        query.execBatch();
    }

    // Finally, we can update the "DB_VERSION" key.
    query.exec(QString::fromUtf8("UPDATE config SET value = 4 WHERE key = 'DB_VERSION'"));

    m_db.commit();
    closeDatabaseConnection();
}

void Database::updateDatabase4to5(QString dbFileName)
{
    connectDatabase(dbFileName);
    QSqlQuery query(m_db);
    m_db.transaction();

    // Get the boogers per round
    query.exec(QString::fromUtf8("SELECT value FROM config WHERE key = 'GAMES_PER_ROUND'"));
    query.next();
    QString boogersPerRound = query.value(0).toString();

    // Update "GAMES_PER_ROUND" to "BOOGERS_PER_ROUND"
    query.exec(QString::fromUtf8("DELETE FROM config WHERE key = 'GAMES_PER_ROUND'"));
    query.prepare(QString::fromUtf8("INSERT INTO config(key, value) VALUES('BOOGERS_PER_ROUND', "
                                                                          "?)"));
    query.addBindValue(boogersPerRound);
    query.exec();

    // Update the score table's "game" column to "booger"
    query.exec(QString::fromUtf8("ALTER TABLE score RENAME TO score_tmp"));
    query.exec(QString::fromUtf8("CREATE TABLE score(\n"
                                 "    round INTEGER,\n"
                                 "    table_number INTEGER,\n"
                                 "    booger INTEGER,\n"
                                 "    pair_or_player INTEGER,\n"
                                 "    score INTEGER\n"
                                 ")"));
    query.exec(QString::fromUtf8("INSERT INTO score(round, table_number, booger, pair_or_player, "
                                                   "score) "
                                 "SELECT round, table_number, game, pair_or_player, score "
                                 "FROM score_tmp"));
    query.exec(QString::fromUtf8("DROP TABLE score_tmp"));

    // Update the "DB_VERSION" key.
    query.exec(QString::fromUtf8("UPDATE config SET value = 5 WHERE key = 'DB_VERSION'"));

    m_db.commit();
    closeDatabaseConnection();
}
