/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATABASE_H
#define DATABASE_H

// Qt includes
#include <QObject>
#include <QSqlDatabase>
#include <QMap>

// Qt classes
class QSqlQuery;

class Database : public QObject
{

public:
    enum OpenError {
        noError,
        tablesMissing,
        version3Db,
        version4Db,
        wrongVersion
    };

    enum TournamentMode {
        fixedPairs,
        singlePlayers
    };

    enum Numerus {
        singular,
        plural
    };

    struct Assignment {
        QMap<int, int> assignedTable;
        QMap<int, int> assignedPairNumber;
        QMap<int, QMap<int, int>> assignedPair;
    };

    struct Score {
        int table;
        QString pairName;
        int score;
    };

    struct SingleScore {
        int pairOrPlayer;
        int score;
    };

    struct Ranking {
        QString name;
        int boogers;
        int goals;
        int opponentGoals;
        int playedBoogers;
    };

    explicit Database(QObject *parent);
    ~Database();
    bool create(QString dbFile);
    bool tournamentStarted() const;
    bool openDb(QString dbFile);
    QString openError() const;
    QString playersString(Numerus numerus, bool article = false) const;
    void reset();
    void saveSettings();
    void deletePlayers(QString name);
    QStringList listPlayers();
    int necessaryDivisor() const;
    bool checkName(QString name);
    void registerPlayers(QString name);
    void registerPlayers(QStringList &names);
    void renamePlayers(QString oldName, QString newName);
    QMap<int, bool> roundsState();
    QList<int> missingTables(int round);
    QMap<QString, int> missingPlayers(int round);
    Assignment getAssignment(int roundNumber);
    QList<Score> getScores(int round);
    QList<SingleScore> getSingleScore(int round, int table);
    QString boogify(int score);
    void saveScore(int round, int table, QList<int> players,
                   QList<int> pair1Scores, QList<int> pair2Scores);
    void deleteScore(int round, int table);
    QList<Ranking> getRanking(int upToRound);
    QString getMuckturnierVersion() const;
    int getDbVersion() const;
    QString getDbFile() const;
    Database::TournamentMode getTournamentMode() const;
    int getBoogerScore() const;
    int getBoogersPerRound() const;
    int playersCount();
    void setTournamentMode(Database::TournamentMode tournamentMode);
    void setBoogerScore(int boogerScore);
    void setBoogersPerRound(int boogersPerRound);
    void closeDatabaseConnection();
    OpenError openErrorCode() const;
    void updateDatabase3to4(QString dbFileName);
    void updateDatabase4to5(QString dbFileName);

private: // Functions
    bool connectDatabase(QString &dbFile);
    void updateSettings();
    void checkIfTournamentStarted();
    void insertScore(QSqlQuery &query,
                     int round, int table, int booger, int pairOrPlayer, int score);
    void updateAssignment(QSqlQuery &query,
                          int round, int pairOrPlayer, int pairNumber, int playerNumber, int table);

private: // Variables
    QSqlDatabase m_db;
    bool m_tournamentStarted;
    QString m_openError;
    QString m_dbValueMuckturnierVersion;
    TournamentMode m_settingsTournamentMode;
    int m_settingsBoogerScore;
    int m_settingsBoogersPerRound;
    OpenError m_openErrorCode;

};

#endif // DATABASE_H
