/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "MainWindow.h"
#include "Database.h"
#include "AboutWindow.h"
#include "PlayersPage.h"
#include "ScorePage.h"
#include "RankingPage.h"
#include "ExportDialog.h"
#include "StopWatch.h"
#include "TournamentInfoDialog.h"
#include "NewTournamentDialog.h"
#include "DefaultValues.h"

// Qt includes
#include <QLabel>
#include <QHBoxLayout>
#include <QTabWidget>
#include <QMenuBar>
#include <QAction>
#include <QApplication>
#include <QStatusBar>
#include <QMessageBox>
#include <QUrl>
#include <QDesktopServices>
#include <QFileInfo>
#include <QDebug>
#include <QFileDialog>
#include <QDir>
#include <QStandardPaths>
#include <QCoreApplication>
#include <QDirIterator>
#include <QDockWidget>
#include <QDesktopWidget>
#include <QPushButton>
#include <QVBoxLayout>
#include <QTextCodec>

namespace {
const QString SPLASHBUTTONSTYLE = QString::fromUtf8("QPushButton { border:none; }");
}

MainWindow::MainWindow()
{
    // Setup the database
    m_db = new Database(this);

    // Window setup
    setWindowIcon(QIcon(resourcePath(QString::fromUtf8("muckturnier.svg"))));
    setWindowTitle(tr("Muckturnier"));

    QRect screenGeometry = QApplication::desktop()->screenGeometry();
    if (screenGeometry.width() >= 1024 && screenGeometry.height() >= 768) {
        resize(900, 700);
    } else {
        setWindowState(Qt::WindowMaximized);
    }

    m_centralWidget = new QWidget(this);
    setCentralWidget(m_centralWidget);
    QHBoxLayout *layout = new QHBoxLayout(m_centralWidget);

    // Startup buttons

    layout->addStretch();
    QVBoxLayout *newTournamentLayout = new QVBoxLayout;
    layout->addLayout(newTournamentLayout);
    newTournamentLayout->addStretch();

    QPushButton *newTournamentButton = new QPushButton;
    newTournamentButton->setStyleSheet(SPLASHBUTTONSTYLE);
    QIcon newIcon = QIcon(resourcePath(QString::fromUtf8("new.svg")));
    newTournamentButton->setIcon(newIcon);
    newTournamentButton->setIconSize(QSize(100, 100));
    connect(newTournamentButton, &QPushButton::clicked, this, &MainWindow::startNewTournament);
    newTournamentLayout->addWidget(newTournamentButton);

    QPushButton *newTournamentLabel = new QPushButton(tr("Ein neues Turnier starten"));
    connect(newTournamentLabel, &QPushButton::clicked, this, &MainWindow::startNewTournament);
    newTournamentLabel->setStyleSheet(SPLASHBUTTONSTYLE);
    newTournamentLayout->addWidget(newTournamentLabel);

    newTournamentLayout->addStretch();
    layout->addStretch();
    QVBoxLayout *loadTournamentLayout = new QVBoxLayout;
    layout->addLayout(loadTournamentLayout);
    loadTournamentLayout->addStretch();

    QPushButton *loadTournamentButton = new QPushButton;
    loadTournamentButton->setStyleSheet(SPLASHBUTTONSTYLE);
    QIcon loadIcon = QIcon(resourcePath(QString::fromUtf8("load.svg")));
    loadTournamentButton->setIcon(loadIcon);
    loadTournamentButton->setIconSize(QSize(100, 100));
    connect(loadTournamentButton, &QPushButton::clicked, this, &MainWindow::triggerLoadTournament);
    loadTournamentLayout->addWidget(loadTournamentButton);

    QPushButton *loadTournamentLabel = new QPushButton(tr("Ein Turnier öffnen"));
    connect(loadTournamentLabel, &QPushButton::clicked, this, &MainWindow::triggerLoadTournament);
    loadTournamentLabel->setStyleSheet(SPLASHBUTTONSTYLE);
    loadTournamentLayout->addWidget(loadTournamentLabel);

    loadTournamentLayout->addStretch();
    layout->addStretch();

    // Dock setup

    setDockNestingEnabled(true);

    m_playersPage = new PlayersPage(this, m_db);
    connect(m_playersPage, &PlayersPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::tournamentCanBeStarted,
            this, &MainWindow::setScoreTabEnabled);
    m_playersDock = new QDockWidget(tr("Paare"));
    m_playersDock->setWidget(m_playersPage);
    formatDockWidget(m_playersDock);
    m_playersDock->hide();

    m_scorePage = new ScorePage(this, m_db);
    connect(m_scorePage, &ScorePage::tournamentStarted, this, &MainWindow::tournamentStarted);
    connect(m_scorePage, &ScorePage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_scorePage, &ScorePage::scoreChanged, m_playersPage, &PlayersPage::scoreChanged);
    m_scoreDock = new QDockWidget(tr("Ergebnisse"));
    m_scoreDock->setWidget(m_scorePage);
    formatDockWidget(m_scoreDock);
    tabifyDockWidget(m_playersDock, m_scoreDock);
    m_scoreDock->hide();

    m_rankingPage = new RankingPage(this, m_db);
    connect(m_rankingPage, &RankingPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::playersListChanged,
            m_rankingPage, &RankingPage::dataChanged);
    connect(m_scorePage, &ScorePage::scoreChanged, m_rankingPage, &RankingPage::dataChanged);
    m_rankingDock = new QDockWidget(tr("Rangliste"));
    m_rankingDock->setWidget(m_rankingPage);
    formatDockWidget(m_rankingDock);
    tabifyDockWidget(m_scoreDock, m_rankingDock);
    m_rankingDock->hide();

    // Menu

    // "File"

    QMenu *fileMenu = menuBar()->addMenu(tr("Datei"));

    QAction *newTournamentAction = new QAction(tr("Neues Turnier starten"), this);
    fileMenu->addAction(newTournamentAction);
    connect(newTournamentAction, &QAction::triggered, this, &MainWindow::startNewTournament);

    QAction *loadTournamentAction = new QAction(tr("Turnier öffnen"), this);
    fileMenu->addAction(loadTournamentAction);
    connect(loadTournamentAction, &QAction::triggered, this, &MainWindow::triggerLoadTournament);

    m_closeTournamentAction = new QAction(tr("Turnier schließen"), this);
    fileMenu->addAction(m_closeTournamentAction);
    m_closeTournamentAction->setEnabled(false);
    connect(m_closeTournamentAction, &QAction::triggered, this, &MainWindow::closeTournament);

    fileMenu->addSeparator();

    m_tournamentInfoAction = new QAction(tr("Turnierinformationen"), this);
    connect(m_tournamentInfoAction, &QAction::triggered, this, &MainWindow::showTournamentInfo);
    fileMenu->addAction(m_tournamentInfoAction);
    m_tournamentInfoAction->setEnabled(false);

    fileMenu->addSeparator();

    QAction *quitAction = new QAction(tr("Beenden"), this);
    fileMenu->addAction(quitAction);
    connect(quitAction, &QAction::triggered, this, &MainWindow::quit);

    // "View"

    QMenu *viewMenu = menuBar()->addMenu(tr("Ansicht"));

    QAction *arrangeTabifiedAction = new QAction(tr("Seiten als Tabs"), this);
    connect(arrangeTabifiedAction, &QAction::triggered, this, &MainWindow::arrangeTabified);
    viewMenu->addAction(arrangeTabifiedAction);

    QAction *arrangeHorizontallyAction = new QAction(tr("Seiten nebeneinander"), this);
    connect(arrangeHorizontallyAction, &QAction::triggered, this, &MainWindow::arrangeHorizontally);
    viewMenu->addAction(arrangeHorizontallyAction);

    QAction *arrangeVerticallyAction = new QAction(tr("Seiten übereinander"), this);
    connect(arrangeVerticallyAction, &QAction::triggered, this, &MainWindow::arrangeVertically);
    viewMenu->addAction(arrangeVerticallyAction);

    QAction *arrangeOwnWindowAction = new QAction(tr("Seiten als einzelne Fenster"), this);
    connect(arrangeOwnWindowAction, &QAction::triggered, this, &MainWindow::arrangeOwnWindow);
    viewMenu->addAction(arrangeOwnWindowAction);

    // "Extras"

    QMenu *extrasMenu = menuBar()->addMenu(tr("Extras"));

    QAction *stopWatchAction = new QAction(tr("Stoppuhr"), this);
    extrasMenu->addAction(stopWatchAction);
    connect(stopWatchAction, &QAction::triggered, this, &MainWindow::showStopWatch);

    extrasMenu->addSeparator();
    m_importPlayersListAction = new QAction(tr("Paarliste importieren"), this);
    connect(m_importPlayersListAction, &QAction::triggered, this, &MainWindow::importPlayersList);
    extrasMenu->addAction(m_importPlayersListAction);
    m_importPlayersListAction->setEnabled(false);

    m_exportPlayersListAction = new QAction(tr("Paarliste exportieren"), this);
    connect(m_playersPage, &PlayersPage::enableExport,
            m_exportPlayersListAction, &QAction::setEnabled);
    connect(m_exportPlayersListAction, &QAction::triggered, this, &MainWindow::exportPlayersList);
    extrasMenu->addAction(m_exportPlayersListAction);
    m_exportPlayersListAction->setEnabled(false);

    m_exportAction = new QAction(tr("Turnierdaten exportieren"), this);
    connect(m_rankingPage, &RankingPage::enableExport, m_exportAction, &QAction::setEnabled);
    connect(m_exportAction, &QAction::triggered, this, &MainWindow::showExportDialog);
    extrasMenu->addAction(m_exportAction);
    m_exportAction->setEnabled(false);

    extrasMenu->addSeparator();

    m_resetDatabaseAction = new QAction(tr("Datenbank zurücksetzen"), this);
    connect(m_playersPage, &PlayersPage::enableExport,
            m_resetDatabaseAction, &QAction::setEnabled);
    extrasMenu->addAction(m_resetDatabaseAction);
    connect(m_resetDatabaseAction, &QAction::triggered, this, &MainWindow::resetDatabase);
    m_resetDatabaseAction->setEnabled(false);

    // "Help"

    QMenu *helpMenu = menuBar()->addMenu(tr("Hilfe"));

    QAction *handbookAction = new QAction(tr("Handbuch zu Muckturnier"), this);
    helpMenu->addAction(handbookAction);
    connect(handbookAction, &QAction::triggered, this, &MainWindow::showHandbook);

    QAction *aboutAction = new QAction(tr("Über Muckturnier"), this);
    helpMenu->addAction(aboutAction);
    connect(aboutAction, &QAction::triggered, this, &MainWindow::showAboutWindow);

    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::startNewTournament()
{
    NewTournamentDialog *dialog = new NewTournamentDialog(this, m_db);
    if (! dialog->exec()) {
        return;
    }

    dbReset();
    databaseChanged();
}

void MainWindow::triggerLoadTournament()
{
    loadTournament();
}

void MainWindow::loadTournament(QString dbToOpen)
{
    QString dbFileName;
    if (! dbToOpen.isEmpty()) {
        dbFileName = dbToOpen;
    } else {
        dbFileName = QFileDialog::getOpenFileName(this,
            tr("Bitte eine Turnierdatenbank auswählen"),
            QDir::toNativeSeparators(QStandardPaths::writableLocation(
                QStandardPaths::DocumentsLocation)),
            tr("Muckturnier-Datenbanken") + QString::fromUtf8(" (*.mtdb *.db);;")
            + tr("Alle Dateien") + QString::fromUtf8(" (*)"));
        if (dbFileName.isEmpty()) {
            return;
        }
    }

    QFileInfo fileInfo = QFileInfo(dbFileName);

    if (! fileInfo.exists()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei konnte nicht gefunden werden.").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! (fileInfo.isFile() || fileInfo.isSymLink())) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Es ist weder eine Datei, noch ein symbolischer Link.").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! fileInfo.isReadable()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei ist nicht lesbar. "
               "Bitte die Zugriffsrechte überprüfen!").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! fileInfo.isWritable()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Laden der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei ist nicht beschreibbar. "
               "Bitte die Zugriffsrechte überprüfen!").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    // At least, we can read and write the file. So let's try to open it

    updateStatus(tr("Öffne Turnierdatenbank …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! m_db->openDb(dbFileName)) {
        QApplication::restoreOverrideCursor();

        Database::OpenError error = m_db->openErrorCode();
        if (error == Database::OpenError::version3Db) {
            updateDatabase(dbFileName, 3);
            return;
        } else if (error == Database::OpenError::version4Db) {
            updateDatabase(dbFileName, 4);
            return;
        }

        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("<p>Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:</p><p>%2</p>").arg(
               QDir::toNativeSeparators(dbFileName), m_db->openError()),
            QMessageBox::Ok);
        return;
    }

    databaseChanged();

    if (m_db->tournamentStarted()) {
        m_scoreDock->raise();
    } else {
        // The score page could be enabled by the previously used database,
        // so disable it first
        m_scoreDock->hide();
        // Re-enable the score page if the database already contains
        // the necessary number of players/pairs
        m_playersPage->reload();
        m_playersDock->raise();
    }

    m_tournamentInfoAction->setEnabled(true);

    updateStatus(tr("Turnierdatenbank %1 geöffnet").arg(QDir::toNativeSeparators(dbFileName)));
    QApplication::restoreOverrideCursor();
}

void MainWindow::updateDatabase(QString dbFileName, int dbVersion)
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Update des Datenbankformats"));
    messageBox.setText(tr("<p><b>Das Datenbankformat von Muckturnier hat sich geändert."
                          "</b></p>"
                          "<p>Soll die Datenbank „%1“ aktualisiert werden? Es wird dabei "
                          "ein Backup der Original-Datenbank im selben Ordner angelegt."
                          "</p>").arg(
                          QDir::toNativeSeparators(dbFileName)));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
    messageBox.setDefaultButton(QMessageBox::Yes);

    if (messageBox.exec() != QMessageBox::Yes) {
        updateStatus(tr("Datenbankupdate abgebrochen!"));
        QMessageBox::critical(this,
            tr("Update des Datenbankformats"),
            tr("Das Update des Datenbankformats wurde abgebrochen. Die Datenbank „%1“ kann nicht "
               "geöffnet werden.").arg(QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Datenbankupdate läuft …"));

    QFileInfo fileInfo(dbFileName);
    QString backupFileName = fileInfo.dir().path() + QDir::separator()
                            + fileInfo.completeBaseName()
                            + QString::fromUtf8("_original_dbv")
                            + QString::number(dbVersion);
    fileInfo.setFile(backupFileName + QString::fromUtf8(".mtdb"));
    int backupNumber = 0;
    while (fileInfo.exists()) {
        backupNumber++;
        fileInfo.setFile(backupFileName + QString::fromUtf8("_")
                         + QString::number(backupNumber)
                         + QString::fromUtf8(".mtdb"));
    }

    QFile backupFile(fileInfo.absoluteFilePath());
    if (! backupFile.open(QIODevice::WriteOnly)) {
        QApplication::restoreOverrideCursor();
        QMessageBox::critical(this,
            tr("Update des Datenbankformats"),
            tr("<p>Es bestehen keine Schreibrechte für die Datei „%1“. Es kann kein Backup der "
               "Original-Datenbank angelegt werden.</p>"
               "<p>Das Update wird abgebrochen. Bitte die Zugriffsrechte überprüfen!</p>").arg(
               QDir::toNativeSeparators(fileInfo.absoluteFilePath())),
            QMessageBox::Ok);
        updateStatus(tr("Datenbankupdate abgebrochen!"));
        return;
    } else {
        backupFile.close();
        backupFile.remove();
        QFile::copy(dbFileName, fileInfo.absoluteFilePath());
    }

    if (dbVersion == 3) {
        m_db->updateDatabase3to4(dbFileName);
        m_db->updateDatabase4to5(dbFileName);
    } else if (dbVersion == 4) {
        m_db->updateDatabase4to5(dbFileName);
    }

    QApplication::restoreOverrideCursor();
    updateStatus(tr("Datenbankupdate beendet!"));
    QMessageBox::information(this,
        tr("Update des Datenbankformats"),
        tr("Die Datenbank „%1“ wurde erfolgreich aktualisiert.").arg(
           QDir::toNativeSeparators(dbFileName)),
        QMessageBox::Ok);

    loadTournament(dbFileName);
}

void MainWindow::quit()
{
    QApplication::quit();
}

void MainWindow::showHandbook()
{
    // Opens the default browser to display the handbook.

    QString readmePath = resourcePath(QString::fromUtf8("Readme.htm"));
    if (! readmePath.isEmpty()) {
        QDesktopServices::openUrl(QUrl::fromLocalFile(readmePath));
        return;
    }

    QMessageBox::critical(this,
                          tr("Handbuch anzeigen"),
                          tr("<p>Das Handbuch kann nicht angezeigt werden: Die Datei „Readme.htm“ "
                             "wurde nicht gefunden.</p>"
                             "<p>Online ist unter <a href=\"http://nasauber.de/opensource/"
                             "muckturnier/readme.php\">http://nasauber.de/opensource/muckturnier/"
                             "readme.php</a> das aktuelle Handbuch zu finden.</p>"),
                          QMessageBox::Ok);
}

void MainWindow::showAboutWindow()
{
    QDialog *aboutWindow = new AboutWindow(this, m_db);
    aboutWindow->show();
}

void MainWindow::updateStatus(QString message)
{
    statusBar()->showMessage(message, 5000);
}

void MainWindow::databaseChanged()
{
    setWindowTitle(tr("Muckturnier: %1").arg(QFileInfo(m_db->getDbFile()).fileName()));
    m_centralWidget->hide();
    setPlayersTabName();
    m_playersDock->show();
    m_playersPage->reload();

    m_exportAction->setEnabled(false);
    m_closeTournamentAction->setEnabled(true);
    m_tournamentInfoAction->setEnabled(true);

    checkEnableRankingPage();
    if (m_db->tournamentStarted()) {
        m_rankingPage->dataChanged();
        m_importPlayersListAction->setEnabled(false);
    } else {
        m_importPlayersListAction->setEnabled(true);
    }

    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        m_exportPlayersListAction->setText(tr("Paarliste exportieren"));
        m_importPlayersListAction->setText(tr("Paarliste importieren"));
        m_importPlayersListString = tr("Paarliste");
        // Extra variable because the action will get a shortcut ampersand somewhere
        m_importPlayersListActionName = tr("Paarliste importieren");
        break;
    case Database::TournamentMode::singlePlayers:
        m_exportPlayersListAction->setText(tr("Spielerliste exportieren"));
        m_importPlayersListAction->setText(tr("Spielerliste importieren"));
        m_importPlayersListString = tr("Spielerliste");
        m_importPlayersListActionName = tr("Spielerliste importieren");
        break;
    }
}

void MainWindow::settingsSaved()
{
    setPlayersTabName();
    m_playersDock->show();
    m_playersPage->reload();
    m_playersDock->raise();
    m_tournamentInfoAction->setEnabled(true);
}

void MainWindow::setScoreTabEnabled(bool state)
{
    m_scoreDock->setVisible(state);
    if(state == true) {
        m_scorePage->reload();
    }
}

void MainWindow::tournamentStarted()
{
    setPlayersTabName();
    m_playersPage->reload();
    checkEnableRankingPage();
    m_importPlayersListAction->setEnabled(false);
}

void MainWindow::setPlayersTabName()
{
    m_playersDock->setWindowTitle(m_db->playersString(Database::Numerus::plural));
}

void MainWindow::checkEnableRankingPage()
{
    m_rankingDock->setVisible(m_db->tournamentStarted());
}

void MainWindow::resizeEvent(QResizeEvent *event)
{
   QMainWindow::resizeEvent(event);
   if (m_scoreDock->isVisible()) {
        m_scorePage->setScrollAreaHeight();
   }
}

void MainWindow::showExportDialog()
{
    ExportDialog *exportDialog = new ExportDialog(this, m_db, m_rankingPage);
    connect(exportDialog, &ExportDialog::statusUpdate, this, &MainWindow::updateStatus);
    exportDialog->exec();
}

void MainWindow::showStopWatch()
{
    StopWatch *stopWatch = new StopWatch(this);
    stopWatch->show();
}

void MainWindow::showTournamentInfo()
{
    TournamentInfoDialog *tournamentInfoDialog = new TournamentInfoDialog(this, m_db);
    tournamentInfoDialog->exec();
}

void MainWindow::formatDockWidget(QDockWidget *dock)
{
    dock->setContextMenuPolicy(Qt::PreventContextMenu);
    dock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);
    dock->setStyleSheet(QString::fromUtf8("QDockWidget { font-weight: bold; font-size: %1pt }").arg(
        (int) double(dock->font().pointSize()) * 1.2));
    addDockWidget(Qt::TopDockWidgetArea, dock);
}

QString MainWindow::resourcePath(QString fileName)
{
    if (m_possibleDataLocations.size() == 0) {
        // Get a list of possible locations for our external files
        QStringList possibleDataLocations;
        const QString iconsPath         = QString::fromUtf8("icons/hicolor/scalable/apps/");
        const QString resourcesPath     = QString::fromUtf8("muckturnier/");
        const QString documentationPath = QString::fromUtf8("doc/muckturnier/");

        // Add the binary's directory (possibly a build directory)
        possibleDataLocations << QCoreApplication::applicationDirPath() + QString::fromUtf8("/");

        // Add a possibly prefixed directory tree
        QUrl prefixed = QUrl(QCoreApplication::applicationDirPath() + QString::fromUtf8("/"))
                            .resolved(QUrl(QString::fromUtf8("../share/")));
        possibleDataLocations << prefixed.toString() + iconsPath
                              << prefixed.toString() + resourcesPath
                              << prefixed.toString() + documentationPath;

        // Add system-wide locations
        for (const QString &path
            : QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation)) {
            possibleDataLocations << path + QString::fromUtf8("/") + iconsPath
                                  << path + QString::fromUtf8("/") + resourcesPath
                                  << path + QString::fromUtf8("/") + documentationPath;
        }

        // Reverse the list
        while (possibleDataLocations.size() > 0) {
            const QString path = QDir::toNativeSeparators(possibleDataLocations.takeLast());
            if (! m_possibleDataLocations.contains(path)) {
                m_possibleDataLocations << path;
            }
        }
    }

    // Check if we already have it
    if (m_resourcePath.contains(fileName)) {
        return m_resourcePath[fileName];
    }

    // First, we search for embedded resources ...
    QDirIterator it(QString::fromUtf8(":"), QDirIterator::Subdirectories);
    while (it.hasNext()) {
        if (it.next() == QString::fromUtf8(":/") + fileName) {
            m_resourcePath[fileName] = QString::fromUtf8(":/") + fileName;
            return m_resourcePath[fileName];
        }
    }

    // ... then we check all other possible data locations
    for (const QString &path : m_possibleDataLocations) {
        QString filePath = path + fileName;
        if (QFileInfo::exists(filePath)) {
            m_resourcePath[fileName] = filePath;
            return m_resourcePath[fileName];
        }
    }

    // This will also return QString() if nothing is found
    return m_resourcePath[fileName];
}

void MainWindow::closeTournament()
{
    m_db->closeDatabaseConnection();
    m_playersDock->hide();
    m_scoreDock->hide();
    m_rankingDock->hide();
    m_centralWidget->show();
    m_closeTournamentAction->setEnabled(false);
    m_tournamentInfoAction->setEnabled(false);
    m_resetDatabaseAction->setEnabled(false);
    m_exportPlayersListAction->setEnabled(false);
    m_exportAction->setEnabled(false);
    m_importPlayersListAction->setEnabled(false);
    setWindowTitle(tr("Muckturnier"));
    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::exportPlayersList()
{
    QString playersListFileName = QFileDialog::getSaveFileName(this,
        tr("Spielerlisten-Export: Bitte einen Dateinamen auswählen"),
        QDir::toNativeSeparators(QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)
                                 + QString::fromUtf8("/Spielerliste.txt")),
        tr("Textdateien") + QString::fromUtf8(" (*.txt);;")
        + tr("Alle Dateien") + QString::fromUtf8(" (*)"));

    if (playersListFileName.isEmpty()) {
        return;
    }

    QFile playersListFile(playersListFileName);
    if (! playersListFile.open(QIODevice::WriteOnly | QIODevice::Text)) {
        QMessageBox::critical(this,
            tr("Turnierdaten-Export"),
            tr("Es bestehen keine Schreibrechte für die Datei „%1“. "
               "Bitte die Zugriffsrechte überprüfen!").arg(
               QDir::toNativeSeparators(playersListFileName)),
            QMessageBox::Ok);
        return;
    }

    QTextStream outStream(&playersListFile);
    outStream.setCodec(QTextCodec::codecForName("UTF-8"));
    updateStatus(tr("Exportiere Spielerliste …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    for (const QString &pairOrPlayer : m_db->listPlayers()) {
        outStream << pairOrPlayer + QString::fromUtf8("\n");
    }

    playersListFile.close();
    QApplication::restoreOverrideCursor();
    updateStatus(tr("Spielerliste in Datei „%1“ exportiert.").arg(
                    QDir::toNativeSeparators(playersListFileName)));

    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Spielerlisten-Export"));
    messageBox.setText(tr("Die Spielerliste wurde erfolgreich in die Datei „%1“ exportiert. "
                          "Soll die Datei jetzt geöffnet werden?").arg(
                          QDir::toNativeSeparators(QDir::toNativeSeparators(playersListFileName))));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::Yes);

    if (messageBox.exec() == QMessageBox::Yes) {
        QDesktopServices::openUrl(QUrl::fromLocalFile(playersListFileName));
    }
}

void MainWindow::attachDocks(Qt::DockWidgetArea area)
{
    m_playersDock->setFloating(false);
    m_scoreDock->setFloating(false);
    m_rankingDock->setFloating(false);
    addDockWidget(area, m_playersDock);
    addDockWidget(area, m_scoreDock);
    addDockWidget(area, m_rankingDock);
}

void MainWindow::arrangeTabified()
{
    attachDocks(Qt::TopDockWidgetArea);
    tabifyDockWidget(m_playersDock, m_scoreDock);
    tabifyDockWidget(m_scoreDock, m_rankingDock);
}

void MainWindow::arrangeHorizontally()
{
    attachDocks(Qt::TopDockWidgetArea);
    splitDockWidget(m_playersDock, m_scoreDock, Qt::Horizontal);
    splitDockWidget(m_scoreDock, m_rankingDock, Qt::Horizontal);
}

void MainWindow::arrangeVertically()
{
    attachDocks(Qt::LeftDockWidgetArea);
    splitDockWidget(m_playersDock, m_scoreDock, Qt::Vertical);
    splitDockWidget(m_scoreDock, m_rankingDock, Qt::Vertical);
}

void MainWindow::arrangeOwnWindow()
{
    m_playersDock->setFloating(true);
    m_scoreDock->setFloating(true);
    m_rankingDock->setFloating(true);
}

void MainWindow::dbReset()
{
    setPlayersTabName();
    m_playersDock->hide();
    m_playersPage->reset();
    m_scoreDock->hide();
    m_rankingDock->hide();
}

void MainWindow::resetDatabase()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Warning);
    messageBox.setWindowTitle(tr("Datenbank zurücksetzen"));
    messageBox.setText(tr("Sollen wirklich alle angemeldeten Spieler und alle eingegebenen "
                          "Ergebnisse gelöscht werden?"));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Setzte Turnierdatenbank zurück …"));

    m_db->reset();
    setPlayersTabName();
    dbReset();
    settingsSaved();

    QApplication::restoreOverrideCursor();
    updateStatus(tr("Turnierdatenbank zurückgesetzt."));
}

void MainWindow::importPlayersList()
{
    // Get the import file's name
    QString inputFileName = QFileDialog::getOpenFileName(this,
        tr("Bitte eine %1 auswählen").arg(m_importPlayersListString),
        QDir::toNativeSeparators(
            QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)),
        tr("Textdateien") + QString::fromUtf8(" (*.txt);;")
        + tr("Alle Dateien") + QString::fromUtf8(" (*)"));
    if (inputFileName.isEmpty()) {
        return;
    }

    // Try to open it
    QFile inputFile(inputFileName);
    if (! inputFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QMessageBox::critical(this,
            m_importPlayersListActionName,
            tr("Es bestehen keine Leserechte für die Datei „%1“. Bitte die Zugriffsrechte "
               "überprüfen!").arg(QDir::toNativeSeparators(inputFileName)),
            QMessageBox::Ok);
        return;
    }

    // Check if it's suspiciously large (more than about 200 mean pair names)
    QFileInfo fileInfo(inputFileName);
    if (fileInfo.size() > IMPORT_SIZE_WARNING) {
        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Warning);
        messageBox.setWindowTitle(m_importPlayersListActionName);
        messageBox.setText(tr("Die Datei „%1“ ist ungewöhnlich groß für eine %1. Soll sie trotzdem "
                              "importiert werden?").arg(QDir::toNativeSeparators(inputFileName),
                                                        m_importPlayersListString));
        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        messageBox.setDefaultButton(QMessageBox::No);

        if (messageBox.exec() == QMessageBox::No) {
            return;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Importiere %1 …").arg(m_importPlayersListString));

    // Get the current player list to search for duplicates
    QStringList allPlayers = m_db->listPlayers();
    for (int i = 0; i < allPlayers.length(); i++) {
        allPlayers[i] = allPlayers[i].toLower();
    }

    QString replacementCharacter = QString::fromUtf8("\xEF\xBF\xBD");
    QStringList playersToAdd;
    int importedPlayers = 0;
    int lineNumber = 0;

    // Read the file line per line

    QTextStream stream(&inputFile);
    stream.setCodec(QTextCodec::codecForName("UTF-8"));

    while (! stream.atEnd()) {
        lineNumber++;
        // Let's truncate lines at byte 256
        // just in case some user opens some binary file or whatnot ;-)
        QString playersName = stream.readLine(256).simplified();
        if (playersName.isEmpty()) {
            continue;
        }

        // If the file contains invalid characters, refuse to import it.
        if (playersName.contains(replacementCharacter)) {
            QString errorLine;
            if (playersName.length() > 45) {
                errorLine = playersName.left(45) + QString::fromUtf8("…");
            } else {
                errorLine = playersName;
            }

            QApplication::restoreOverrideCursor();
            QMessageBox::critical(this,
                m_importPlayersListActionName,
                tr("<p>Die Datei „%1“ enthält ungültige Zeichen. Vermutlich ist es keine reine "
                   "Textdatei, oder sie ist nicht UTF-8-kodiert. Zeile Nr. %2 war die erste "
                   "fehlerhalt kodierte:</p><p>%3</p>"
                   "<p>Es kann eine UTF-8-kodierte Textdatei importiert werden, die pro Zeile "
                   "einen Paar- bzw. Spielernamen enthält.</p>").arg(
                   QDir::toNativeSeparators(inputFileName), QString::number(lineNumber), errorLine),
                   QMessageBox::Ok);
            return;
        }

        // Check for suspiciously long lines
        if (playersName.length() > IMPORT_LINE_LENGTH_WARNING) {
            QApplication::restoreOverrideCursor();
            QMessageBox messageBox(this);
            messageBox.setIcon(QMessageBox::Warning);
            messageBox.setWindowTitle(m_importPlayersListActionName);
            messageBox.setText(tr("Die Datei „%1“ enthält ungewöhnlich lange Zeilen. Soll sie "
                                  "trotzdem importiert werden?").arg(
                                  QDir::toNativeSeparators(inputFileName)));
            messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
            messageBox.setDefaultButton(QMessageBox::No);

            if (messageBox.exec() == QMessageBox::No) {
                return;
            }
        }

        // Check for a duplicate existing name and generate non-colliding names.
        int i = 1;
        QString checkName = playersName.toLower();
        while (true) {
            bool found = false;
            for (const QString &name : allPlayers) {
                if (name == checkName) {
                    found = true;
                    break;
                }
            }
            if (! found) {
                break;
            } else {
                checkName = playersName.toLower() + QString::fromUtf8(" (%1)").arg(++i);
            }
        }
        if (i > 1) {
            playersName = playersName + QString::fromUtf8(" (%1)").arg(i);
        }

        allPlayers << playersName.toLower();
        playersToAdd << playersName;
        importedPlayers++;
    }

    inputFile.close();

    // Check for suspiciously many pairs/players
    if (importedPlayers > IMPORT_LINES_COUNT_WARNING) {
        QApplication::restoreOverrideCursor();
        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Warning);
        messageBox.setWindowTitle(m_importPlayersListActionName);
        messageBox.setText(tr("Die Datei „%1“ enthält %2 Zeilen! Soll sie wirklich importiert "
                              "werden?").arg(QDir::toNativeSeparators(inputFileName),
                                             QString::number(importedPlayers)));
        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        messageBox.setDefaultButton(QMessageBox::No);

        if (messageBox.exec() == QMessageBox::No) {
            return;
        }
    }

    // Do the actual import
    if (importedPlayers > 0) {
        m_db->registerPlayers(playersToAdd);
        m_playersPage->reload();
        m_exportPlayersListAction->setEnabled(true);
        m_resetDatabaseAction->setEnabled(true);
    }

    QApplication::restoreOverrideCursor();
    updateStatus(tr("%1 importiert.").arg(m_importPlayersListString));
    QMessageBox::information(this,
        m_importPlayersListActionName,
        tr("Es wurden %1 %2 importiert!").arg(
            QString::number(importedPlayers),
            m_db->playersString((importedPlayers == 1) ? Database::Numerus::singular
                                                       : Database::Numerus::plural)),
        QMessageBox::Ok);
}
