/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "NewTournamentDialog.h"
#include "Database.h"
#include "MainWindow.h"
#include "DefaultValues.h"

// Qt includes
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QApplication>
#include <QStyle>
#include <QDebug>
#include <QFileDialog>
#include <QStandardPaths>
#include <QDialogButtonBox>
#include <QDir>
#include <QRadioButton>
#include <QGridLayout>
#include <QSpinBox>
#include <QMessageBox>
#include <QFileInfo>

NewTournamentDialog::NewTournamentDialog(QWidget *parent, Database *db)
    : QDialog(parent), m_db(db)
{
    setAttribute(Qt::WA_DeleteOnClose);
    setWindowTitle(tr("Muckturnier: Neues Turnier starten"));

    QVBoxLayout *layout = new QVBoxLayout;
    setLayout(layout);

    m_fileName = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)
                 + QString::fromUtf8("/Muckturnier.mtdb");
    int number = 0;
    while (QFileInfo::exists(m_fileName)) {
        number++;
        m_fileName = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)
                     + QString::fromUtf8("/Muckturnier_%1.mtdb").arg(number);
    }

    // Database filename

    QGroupBox *filenameBox = new QGroupBox(tr("Turnierdatenbank-Dateiname"));
    layout->addWidget(filenameBox);
    QHBoxLayout *fileNameBoxLayout = new QHBoxLayout(filenameBox);

    m_fileNameLabel = new QLabel(QDir::toNativeSeparators(m_fileName));
    m_fileNameLabel->setWordWrap(true);
    fileNameBoxLayout->addWidget(m_fileNameLabel);

    QPushButton *chooseFileNameButton = new QPushButton;
    chooseFileNameButton->setToolTip(tr("Einen Dateinamen auswählen/eingeben"));
    chooseFileNameButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogOpenButton));
    chooseFileNameButton->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    fileNameBoxLayout->addWidget(chooseFileNameButton);
    connect(chooseFileNameButton, &QPushButton::clicked,
            this, &NewTournamentDialog::chooseFileName);

    // Tournament type

    QGroupBox *tournamentTypeBox = new QGroupBox(tr("Turniertyp"));
    QGridLayout *tournamentTypeBoxLayout = new QGridLayout(tournamentTypeBox);
    layout->addWidget(tournamentTypeBox);

    m_fixedPairs = new QRadioButton(tr("Feste Paare"));
    m_fixedPairs->setChecked(Qt::Checked);
    tournamentTypeBoxLayout->addWidget(m_fixedPairs, 0, 0, Qt::AlignTop);

    QLabel *fixedPairsDescription = new QLabel(tr("Es werden immer zwei Spieler als festes Paar "
                                                  "angemeldet, das das ganze Turnier zusammen "
                                                  "spielt."));
    fixedPairsDescription->setWordWrap(true);
    tournamentTypeBoxLayout->addWidget(fixedPairsDescription, 0, 1);

    m_singlePlayers = new QRadioButton(tr("Einzelne Spieler"));
    tournamentTypeBoxLayout->addWidget(m_singlePlayers, 1, 0, Qt::AlignTop);

    QLabel *singlePlayersDescription = new QLabel(tr("Es werden einzelne Spieler angemeldet. Die "
                                                     "Paare werden pro Runde ausgelost."));
    singlePlayersDescription->setWordWrap(true);
    tournamentTypeBoxLayout->addWidget(singlePlayersDescription, 1, 1);

    // Tournament settings

    QGroupBox *tournamentSettingsBox = new QGroupBox(tr("Turniereinstellungen"));
    QHBoxLayout *tournamentSettingsBoxWrapperLayout = new QHBoxLayout(tournamentSettingsBox);
    QGridLayout *tournamentSettingsBoxLayout = new QGridLayout;
    tournamentSettingsBoxWrapperLayout->addLayout(tournamentSettingsBoxLayout);
    tournamentSettingsBoxWrapperLayout->addStretch();
    layout->addWidget(tournamentSettingsBox);

    QLabel *boogerScoreLabel = new QLabel(tr("Punkte pro Bobbl:"));
    boogerScoreLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentSettingsBoxLayout->addWidget(boogerScoreLabel, 0, 0);

    m_boogerScore = new QSpinBox;
    m_boogerScore->setMinimum(1);
    m_boogerScore->setMaximum(MAXIMUM_BOOGER_SCORE);
    m_boogerScore->setValue(DEFAULT_BOOGER_SCORE);
    m_boogerScore->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentSettingsBoxLayout->addWidget(m_boogerScore, 0, 1);

    QLabel *boogersPerRoundLabel = new QLabel(tr("Bobbl pro Runde:"));
    boogersPerRoundLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentSettingsBoxLayout->addWidget(boogersPerRoundLabel, 1, 0);

    m_boogersPerRound = new QSpinBox;
    m_boogersPerRound->setMinimum(1);
    m_boogersPerRound->setMaximum(MAXIMUM_BOOGERS_PER_ROUND);
    m_boogersPerRound->setValue(DEFAULT_BOOGERS_PER_ROUND);
    m_boogersPerRound->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentSettingsBoxLayout->addWidget(m_boogersPerRound, 1, 1);

    // Dialog buttons

    m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);

    m_buttonBox->button(QDialogButtonBox::Ok)->setText(tr("Turnier starten"));
    connect(m_buttonBox->button(QDialogButtonBox::Ok), &QPushButton::clicked,
            this, &NewTournamentDialog::createNewTournament);

    m_buttonBox->button(QDialogButtonBox::Cancel)->setText(tr("Abbrechen"));
    connect(m_buttonBox->button(QDialogButtonBox::Cancel), &QPushButton::clicked,
            this, &QDialog::reject);

    layout->addWidget(m_buttonBox);
}

void NewTournamentDialog::chooseFileName()
{
    QString fileName = QFileDialog::getSaveFileName(this,
        tr("Bitte einen Dateinamen für die Turnierdatenbank auswählen"),
        QDir::toNativeSeparators(m_fileName),
        tr("Muckturnier-Datenbanken") + QString::fromUtf8(" (*.mtdb);;")
        + tr("Alle Dateien") + QString::fromUtf8(" (*)"));

    if (fileName.isEmpty()) {
        m_fileNameLabel->setText(QString::fromUtf8("<span style=\"color:red;\">")
                                 + tr("Bitte einen Dateinamen auswählen!")
                                 + QString::fromUtf8("</span>"));
        m_buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
    } else {
        m_fileNameLabel->setText(QDir::toNativeSeparators(fileName));
        m_buttonBox->button(QDialogButtonBox::Ok)->setEnabled(true);
        m_fileName = fileName;
    }
}

void NewTournamentDialog::createNewTournament()
{
    // Yes, we have to do it in this way, because QFileInfo::isWritable
    // does not work as expected on Windows (returns false for "My Documents").
    QFile test(m_fileName);
    if (test.open(QIODevice::WriteOnly)) {
        test.close();
        test.remove();
    } else {
        QMessageBox::critical(this,
            tr("Neues Turnier starten"),
            tr("Es bestehen keine Schreibrechte für die Datei „%1“. "
               "Bitte die Zugriffsrechte überprüfen!").arg(QDir::toNativeSeparators(m_fileName)),
            QMessageBox::Ok);
        return;
    }

    MainWindow *mainWindow = dynamic_cast<MainWindow*>(parent());

    mainWindow->updateStatus(tr("Lege neue Turnierdatenbank an …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! m_db->create(m_fileName)) {
        // This should not happen
        mainWindow->updateStatus(tr("Das Anlegen der Turnierdatenbank „%1“ ist "
                                    "fehlgeschlagen!").arg(
                                    QDir::toNativeSeparators(m_fileName)));
        QApplication::restoreOverrideCursor();
        return;
    }

    if (m_fixedPairs->isChecked()) {
        m_db->setTournamentMode(Database::TournamentMode::fixedPairs);
    } else {
        m_db->setTournamentMode(Database::TournamentMode::singlePlayers);
    }
    m_db->setBoogerScore(m_boogerScore->value());
    m_db->setBoogersPerRound(m_boogersPerRound->value());
    m_db->saveSettings();

    QApplication::restoreOverrideCursor();
    mainWindow->updateStatus(tr("Turnier „%1“ gestartet.").arg(
                                QDir::toNativeSeparators(m_fileName)));
    QDialog::accept();
}
