/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "RankingPage.h"
#include "Database.h"
#include "Cell.h"
#include "version.h"
#include "SearchWidget.h"
#include "GridDelegate.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QTableWidget>
#include <QHeaderView>
#include <QCheckBox>
#include <QComboBox>

// C++ includes
#include <algorithm> // std::sort

RankingPage::RankingPage(QWidget *parent, Database *db)
    : QWidget(parent), m_db(db)
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    QGroupBox *settingsBox = new QGroupBox;
    QHBoxLayout *settingsBoxLayout = new QHBoxLayout(settingsBox);
    layout->addWidget(settingsBox);

    m_regardOpponentGoals = new QCheckBox(tr("Gegnerische Tore berücksichtigen"));
    m_regardOpponentGoals->setToolTip(tr(
        "Wenn diese Option ausgewählt ist, dann werden bei der Erstellung\n"
        "der Rangliste die geschossenen Tore der Gegner in den gewonnenen\n"
        "Bobbeln berücksichtigt – je weniger, desto besser."
    ));
    connect(m_regardOpponentGoals, &QCheckBox::clicked, this, &RankingPage::toggleOpponentGoals);
    settingsBoxLayout->addWidget(m_regardOpponentGoals);

    m_upToRoundLabel = new QLabel(tr("Rangliste anzeigen für Runde"));
    m_upToRoundLabel->setAlignment(Qt::AlignRight);
    settingsBoxLayout->addWidget(m_upToRoundLabel);

    m_upToRound = new QComboBox;
    m_upToRound->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    m_upToRound->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    connect(m_upToRound, static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged),
            this, &RankingPage::showRankingUpToRound);
    settingsBoxLayout->addWidget(m_upToRound);

    m_rankingBox = new QGroupBox;
    QVBoxLayout *rankingBoxLayout = new QVBoxLayout(m_rankingBox);
    layout->addWidget(m_rankingBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    rankingBoxLayout->addWidget(m_isSearchLabel);

    m_interimLabel = new QLabel;
    m_interimLabel->setWordWrap(true);
    rankingBoxLayout->addWidget(m_interimLabel);
    m_interimLabel->hide();

    m_ranking = new QTableWidget;
    m_ranking->setItemDelegate(new GridDelegate(this));
    m_ranking->verticalHeader()->hide();
    m_ranking->setShowGrid(false);
    m_ranking->setWordWrap(false);
    m_ranking->setSelectionMode(QTableWidget::NoSelection);
    m_ranking->setFocusPolicy(Qt::NoFocus);
    m_ranking->setColumnCount(4);
    rankingBoxLayout->addWidget(m_ranking);

    m_multipleRanks = new QLabel;
    m_multipleRanks->setWordWrap(true);
    m_multipleRanks->hide();
    rankingBoxLayout->addWidget(m_multipleRanks);

    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &RankingPage::updateSearch);
    rankingBoxLayout->addWidget(m_searchWidget);
}

void RankingPage::dataChanged()
{
    m_roundsState = m_db->roundsState();
    m_allRounds = m_roundsState.size();

    m_upToRound->blockSignals(true);
    m_upToRound->clear();
    for (int i = 1; i <= m_allRounds; i++) {
        m_upToRound->addItem(QString::number(i));
    }
    m_upToRound->setCurrentIndex(m_allRounds - 1);
    m_upToRound->blockSignals(false);

    if (m_upToRound->count() == 1) {
        m_upToRound->hide();
        m_upToRoundLabel->hide();
    } else {
        m_upToRound->show();
        m_upToRoundLabel->show();
    }

    reload(m_allRounds);
}

void RankingPage::toggleOpponentGoals()
{
    reload(m_upToRound->currentIndex() + 1);
}

void RankingPage::showRankingUpToRound(int round)
{
    reload(round + 1);
}

void RankingPage::reload(int upToRound)
{
    m_searchWidget->saveSearch();

    QList<int> incompleteRounds;
    QList<int> roundNumbers = m_roundsState.keys();
    std::sort(roundNumbers.begin(), roundNumbers.end());
    for (const int &round : roundNumbers) {
        if (! m_roundsState[round]) {
            incompleteRounds << round;
        }
    }

    int missingRounds = incompleteRounds.size();
    if (missingRounds != 0) {
        emit enableExport(false);
    } else {
        emit enableExport(true);
    }

    QString interimText;
    if (missingRounds > 0) {
        interimText = tr("<table><tr><td><b>Achtung!</b></td><td>Das ist nur ein Zwischenergebnis. "
                         "Es fehlen noch Ergebnisse aus ");
        if (missingRounds == 1) {
            interimText += tr("der Runde %1.").arg(QString::number(incompleteRounds[0]));
        } else if (missingRounds == 2) {
            interimText += tr("den Runden %1 und %2.").arg(QString::number(incompleteRounds[0]),
                                                        QString::number(incompleteRounds[1]));
        } else {
            interimText += tr("den Runden ");
            for (int i = 0; i < incompleteRounds.size() - 1; i++) {
                interimText += tr("%1, ").arg(QString::number(incompleteRounds[i]));
            }
            interimText = interimText.left(interimText.size() - 2);
            interimText += tr(" und %1.").arg(QString::number(incompleteRounds.last()));
        }
        interimText += tr("<br/>%1, von denen noch Spielstände fehlen, sind "
                          "<span style=\"color:red\">rot</span> markiert.</td></tr></table>").arg(
                          m_db->playersString(Database::Numerus::plural));
    }

    if (upToRound < m_allRounds) {
        if (m_roundsState[upToRound]) {
            m_rankingBox->setTitle(tr("Zwischenstand: Rangliste nach Runde %1").arg(upToRound));
        } else {
            m_rankingBox->setTitle(tr("Zwischenstand: Rangliste im Verlauf von Runde %1").arg(
                                      upToRound));
        }
        interimText = tr("<table><tr><td><b>Achtung!</b></td><td>Das ist nicht die aktuelle "
                         "Rangliste! Um diese anzuzeigen <b>bitte Runde %1 auswählen!</b></td>"
                         "</tr></table>").arg(m_allRounds)
                      + interimText;
    } else {
        if (! m_roundsState[m_allRounds]) {
            m_rankingBox->setTitle(tr("Aktuelle Rangliste im Verlauf der %1. Runde").arg(
                                      QString::number(m_allRounds)));
        } else {
            if (m_allRounds == 1) {
                m_rankingBox->setTitle(tr("Aktuelle Rangliste nach einer Runde"));
            } else {
                m_rankingBox->setTitle(tr("Aktuelle Rangliste nach %1 Runden").arg(
                                          QString::number(m_allRounds)));
            }
        }
    }

    if (interimText.isEmpty()) {
        m_interimLabel->hide();
    } else {
        m_interimLabel->setText(interimText);
        m_interimLabel->show();
    }

    QList<Database::Ranking> allRankings = m_db->getRanking(upToRound);
    int allBoogers = upToRound * m_db->getBoogersPerRound();

    m_ranking->clearContents();
    m_ranking->setRowCount(0);
    if (m_regardOpponentGoals->isChecked()) {
        m_ranking->setColumnCount(5);
        m_ranking->setHorizontalHeaderLabels(QStringList()
            << tr("Platz")
            << m_db->playersString(Database::Numerus::singular)
            << tr("Bobbl")
            << tr("Tore")
            << tr("Gegn. Tore"));
    } else {
        m_ranking->setColumnCount(4);
        m_ranking->setHorizontalHeaderLabels(QStringList()
            << tr("Platz")
            << m_db->playersString(Database::Numerus::singular)
            << tr("Bobbl")
            << tr("Tore"));
    }

    m_ranking->setRowCount(allRankings.size());
    int rank = 0;
    int trueRank = 0;
    int prevRank = 0;
    int prevBoogers = 0;
    int prevGoals = 0;
    int prevOpponentGoals = 0;

    int color = 1;
    bool rankDiffers = false;
    QList<int> multipleRanks;
    QMap<int, int> multipleRanksCount;
    for (const Database::Ranking &ranking : allRankings) {
        trueRank++;
        if (m_regardOpponentGoals->isChecked()) {
            rankDiffers = ranking.boogers != prevBoogers
                          || ranking.goals != prevGoals
                          || ranking.opponentGoals != prevOpponentGoals;
        } else {
            rankDiffers = ranking.boogers != prevBoogers
                          || ranking.goals != prevGoals;
        }
        if (rankDiffers) {
            rank = trueRank;
        } else {
            rank = prevRank;
        }

        if (rank != prevRank) {
            color++;
            color = color % 2;
        } else {
            multipleRanks << rank;
            multipleRanksCount[rank]++;
        }

        prevRank = rank;
        prevBoogers = ranking.boogers;
        prevGoals = ranking.goals;
        prevOpponentGoals = ranking.opponentGoals;

        Qt::GlobalColor textColor = Qt::color0;
        if (ranking.playedBoogers != allBoogers) {
            textColor = Qt::red;
        }

        m_ranking->setItem(
            trueRank - 1, 0, new Cell(QString::number(rank), false, color, textColor));
        m_ranking->setItem(
            trueRank - 1, 1, new Cell(ranking.name, false, color, textColor));
        m_ranking->setItem(
            trueRank - 1, 2, new Cell(QString::number(ranking.boogers), false, color, textColor));
        m_ranking->setItem(
            trueRank - 1, 3, new Cell(QString::number(ranking.goals), false, color, textColor));
        if (m_regardOpponentGoals->isChecked()) {
            m_ranking->setItem(trueRank - 1, 4,
                               new Cell(QString::number(ranking.opponentGoals),
                                        false, color, textColor));
        }
    }

    if (multipleRanks.count() > 0) {
        QString multipleRanksText = tr("Mehrfach besetzte Ranglistenplätze: ");
        for (const int &rank : multipleRanks) {
            multipleRanksText += tr("<b>%1</b> (%2x), ").arg(
                                    QString::number(rank),
                                    QString::number(multipleRanksCount[rank] + 1));
        }
        multipleRanksText.chop(2);
        m_multipleRanks->setText(multipleRanksText);
        m_multipleRanks->show();
    } else {
        m_multipleRanks->hide();
    }

    m_ranking->resizeColumnsToContents();
    m_searchWidget->restoreSearch();
}

void RankingPage::updateSearch(QString searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_ranking->rowCount(); row++) {
            if (m_ranking->isRowHidden(row)) {
                m_ranking->showRow(row);
            }
        }
        m_ranking->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    for (int row = 0; row < m_ranking->rowCount(); row++) {
        if (m_ranking->item(row, 1)->text().contains(searchText, Qt::CaseInsensitive)) {
            if (m_ranking->isRowHidden(row)) {
                m_ranking->showRow(row);
            }
            matches++;
        } else {
            if (! m_ranking->isRowHidden(row)) {
                m_ranking->hideRow(row);
            }
        }
    }
    m_searchWidget->updateMatches(matches);
    m_ranking->resizeColumnsToContents();

    if (! m_isSearchLabel->isVisible()) {
        m_isSearchLabel->show();
    }
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
}
