/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ScorePage.h"
#include "BoogerWidget.h"
#include "Cell.h"
#include "SearchWidget.h"
#include "GridDelegate.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QComboBox>
#include <QGridLayout>
#include <QCheckBox>
#include <QScrollArea>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QMenu>
#include <QAction>
#include <QHBoxLayout>
#include <QMessageBox>
#include <QApplication>
#include <QScrollBar>
#include <QWidgetAction>
#include <QLabel>

ScorePage::ScorePage(QWidget *parent, Database *db)
    : QWidget(parent), m_db(db)
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Round

    QGroupBox *roundBox = new QGroupBox;
    QHBoxLayout *roundBoxLayout = new QHBoxLayout(roundBox);
    layout->addWidget(roundBox);

    QLabel *roundLabel = new QLabel(QString::fromUtf8("<b>") + tr("Runde:")
                                    + QString::fromUtf8("</b>"));
    roundLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    roundBoxLayout->addWidget(roundLabel);

    m_roundSelect = new QComboBox;
    m_roundSelect->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    m_roundSelect->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    roundBoxLayout->addWidget(m_roundSelect);
    connect(m_roundSelect, static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged),
            this, &ScorePage::openRound);

    m_abortRound = new QCheckBox(tr("Die Runde wurde abgebrochen"));
    m_abortRound->setToolTip(tr("Wenn diese Option aktiviert ist, können Bobbl als abgebrochen\n"
                                "markiert werden. In diesem Fall können für beide Teams Punkte\n"
                                "eingegeben werden, die jeweils als „geschossene Tore“ zählen."));
    roundBoxLayout->addWidget(m_abortRound);

    roundBoxLayout->addStretch();

    m_resultBox = new QGroupBox(tr("Ergebnisse eintragen"));
    QVBoxLayout *resultBoxLayout = new QVBoxLayout(m_resultBox);
    layout->addWidget(m_resultBox);

    // Table and pairs/players

    QGroupBox *playersBox = new QGroupBox;
    QGridLayout *playersBoxLayout = new QGridLayout(playersBox);
    resultBoxLayout->addWidget(playersBox);

    QLabel *tableLabel = new QLabel(QString::fromUtf8("<b>") + tr("Tisch:")
                                    + QString::fromUtf8("</b>"));
    tableLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    playersBoxLayout->addWidget(tableLabel, 0, 0, 2, 1);

    m_tableSelect = new QComboBox;
    m_tableSelect->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    m_tableSelect->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    playersBoxLayout->addWidget(m_tableSelect, 0, 1, 2, 1);
    connect(m_tableSelect, static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged),
            this, &ScorePage::tableChanged);

    for (int players = 1; players <= 2; players++) {
        QLabel *playersLabel = new QLabel(QString::fromUtf8("<b>")
                                          + tr("Paar %1:").arg(QString::number(players))
                                          + QString::fromUtf8("</b>"));
        playersLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
        playersBoxLayout->addWidget(playersLabel, players - 1, 2);

        QMap<int, QComboBox *> entry;
        for (int player = 1; player <= 2; player++) {
            QComboBox *playerComboBox = new QComboBox;
            playerComboBox->setSizeAdjustPolicy(QComboBox::AdjustToContents);
            playerComboBox->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
            playersBoxLayout->addWidget(playerComboBox, players - 1, 2 + player);
            entry[player] = playerComboBox;
        }

        m_playersSelect[players] = entry;
    }

    connect(m_playersSelect[1][1],
            static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged),
            this, &ScorePage::pair1Changed);
    connect(m_playersSelect[2][1],
            static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged),
            this, &ScorePage::pair2Changed);
    connect(m_playersSelect[1][2],
            static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged),
            this, &ScorePage::pair34Changed);
    connect(m_playersSelect[2][2],
            static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged),
            this, &ScorePage::pair34Changed);

    m_autoSelect = new QCheckBox(tr("Tischnummer bzw. Paare\nautomatisch auswählen"));
    m_autoSelect->setChecked(Qt::Checked);
    playersBoxLayout->addWidget(m_autoSelect, 0, 5, 2, 1);

    // This works as a spacer so that the place is filled if m_autoSelect is not shown
    playersBoxLayout->addWidget(new QLabel, 0, 6);

    // Enter Score
    QGroupBox *scoreBox = new QGroupBox;
    QVBoxLayout *scoreBoxLayout = new QVBoxLayout(scoreBox);
    scoreBox->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    resultBoxLayout->addWidget(scoreBox);

    m_scoreBoxScroll = new QScrollArea(scoreBox);
    m_scoreBoxScroll->setFrameShape(QFrame::NoFrame);
    scoreBoxLayout->addWidget(m_scoreBoxScroll);

    m_scoreBoxWidget = new QWidget;

    m_saveScoreButton = new QPushButton(tr("Ergebnis\neintragen"));
    connect(m_saveScoreButton, &QPushButton::clicked, this, &ScorePage::saveScore);

    // The round's results

    m_roundSummaryBox = new QGroupBox;
    QVBoxLayout *roundSummaryBoxLayout = new QVBoxLayout(m_roundSummaryBox);
    layout->addWidget(m_roundSummaryBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    roundSummaryBoxLayout->addWidget(m_isSearchLabel);

    m_scoreList = new QTableWidget;
    m_scoreList->setItemDelegate(new GridDelegate(this, GridDelegate::HighlightMode::evenRows));
    m_scoreList->verticalHeader()->hide();
    m_scoreList->setShowGrid(false);
    m_scoreList->setWordWrap(false);
    m_scoreList->setSelectionMode(QTableWidget::NoSelection);
    m_scoreList->setFocusPolicy(Qt::NoFocus);
    m_scoreList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_scoreList, &QTableWidget::cellDoubleClicked,
            this, &ScorePage::triggerShowContextMenu);
    connect(m_scoreList, &QTableWidget::customContextMenuRequested,
            this, &ScorePage::showContextMenu);
    roundSummaryBoxLayout->addWidget(m_scoreList);

    // Search widget

    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &ScorePage::updateSearch);
    roundSummaryBoxLayout->addWidget(m_searchWidget);

    // Context menu

    m_contextMenu = new QMenu(this);

    m_menuTitle = new QLabel;
    m_menuTitle->setAlignment(Qt::AlignCenter);
    m_menuTitle->setMargin(4);
    QWidgetAction *titleAction = new QWidgetAction(m_contextMenu);
    titleAction->setDefaultWidget(m_menuTitle);
    m_contextMenu->addAction(titleAction);
    m_contextMenu->addSeparator();

    m_editAction = new QAction(tr("Ändern (Löschen und neu eingeben)"), this);
    m_contextMenu->addAction(m_editAction);
    connect(m_editAction, &QAction::triggered, this, &ScorePage::editScore);

    m_swapPairsAction = new QAction(tr("Paar 1 und Paar 2 vertauschen"), this);
    m_contextMenu->addAction(m_swapPairsAction);
    connect(m_swapPairsAction, &QAction::triggered, this, &ScorePage::swapPairs);

    m_contextMenu->addSeparator();

    m_deleteAction = new QAction(tr("Löschen"), this);
    m_contextMenu->addAction(m_deleteAction);
    connect(m_deleteAction, &QAction::triggered, this, &ScorePage::deleteScore);
}

void ScorePage::reload(int lastRound, int abortRound)
{
    blockSelectSignals(true);
    m_saveScoreButton->setEnabled(false);

    // Show the needed input fields for pairs or players
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        m_playersSelect[1][2]->hide();
        m_playersSelect[2][2]->hide();
        m_autoSelect->show();
        break;
    case Database::TournamentMode::singlePlayers:
        m_playersSelect[1][2]->show();
        m_playersSelect[2][2]->show();
        m_autoSelect->setChecked(false);
        m_autoSelect->hide();
        break;
    }

    // Display the correct number of booger widgets

    m_boogerWidgets = QList<BoogerWidget *>();
    m_scoreBoxWidget->deleteLater();
    m_scoreBoxWidget = new QWidget;
    m_scoreBoxWidgetLayout = new QHBoxLayout(m_scoreBoxWidget);
    m_scoreBoxWidgetLayout->setSizeConstraint(QLayout::SetFixedSize);

    int boogersPerRound = m_db->getBoogersPerRound();
    for (int i = 1; i <= boogersPerRound; i++) {
        m_boogerWidgets << new BoogerWidget(m_scoreBoxWidget, i, boogersPerRound,
                                            m_db->getBoogerScore());
        m_scoreBoxWidgetLayout->addWidget(m_boogerWidgets.last());
        connect(m_abortRound, &QCheckBox::stateChanged,
                m_boogerWidgets.last(), &BoogerWidget::abortRound);
    }
    connect(m_boogerWidgets.last(), &BoogerWidget::enterPressed, this, &ScorePage::checkEnterScore);

    // Add the "Save score" button
    m_scoreBoxWidgetLayout->addWidget(m_saveScoreButton);
    m_saveScoreButton->setEnabled(false);
    m_scoreBoxScroll->setWidget(m_scoreBoxWidget);
    m_scoreBoxWidget->setAutoFillBackground(false);
    setScrollAreaHeight();

    // Fill the round select

    int allRounds = 0;
    int firstOpenRound = -1;

    m_roundSelect->clear();
    m_roundsState = m_db->roundsState();

    QMapIterator<int, bool> roundState(m_roundsState);
    while (roundState.hasNext()) {
        roundState.next();
        if (roundState.value()) {
            m_roundSelect->addItem(tr("%1 (abgeschlossen)").arg(
                                      QString::number(roundState.key())), roundState.key());
        } else {
            m_roundSelect->addItem(tr("%1 (offen)").arg(
                                      QString::number(roundState.key())), roundState.key());
            if (firstOpenRound == -1) {
                firstOpenRound = m_roundSelect->count() - 1;
            }
        }
        allRounds++;
    }

    if (firstOpenRound == -1) {
        firstOpenRound = 0;
    }

    if (lastRound != -1) {
        firstOpenRound = lastRound - 1;
    }

    // Add one for a new round
    allRounds++;
    m_roundSelect->addItem(tr("%1 (neue Runde)").arg(QString::number(allRounds)), allRounds);
    m_roundsState[allRounds] = false;

    m_roundSelect->setCurrentIndex(firstOpenRound);
    openRound(firstOpenRound);
    blockSelectSignals(false);

    // Update the score widget
    m_scoreList->setColumnCount(m_db->getBoogersPerRound() + 2);

    QStringList headerLabels;
    headerLabels << tr("Tisch") << tr("Paare");
    for (int i = 1; i <= m_db->getBoogersPerRound(); i++) {
        headerLabels << tr("%1. Bobbl").arg(QString::number(i));
    }
    headerLabels << QString();

    m_scoreList->setHorizontalHeaderLabels(headerLabels);
    updateScore();

    m_abortRound->setChecked(abortRound);
}

void ScorePage::setScrollAreaHeight()
{
    if (m_scoreBoxScroll->width() < m_scoreBoxWidget->width()) {
        m_scoreBoxScroll->setMinimumHeight(m_scoreBoxWidget->height()
                                           + m_scoreBoxScroll->horizontalScrollBar()->height());
    } else {
        m_scoreBoxScroll->setMinimumHeight(m_scoreBoxWidget->height());
    }
}

void ScorePage::openRound(int roundNumber)
{
    if (roundNumber == -1) {
        return;
    }

    m_searchWidget->saveSearch();
    blockSelectSignals(true);

    int round = m_roundSelect->itemData(roundNumber).toInt();

    if (m_roundsState.value(round)) {
        m_resultBox->hide();
        m_abortRound->hide();
    } else {
        m_resultBox->show();
        m_abortRound->show();
        m_abortRound->setChecked(Qt::Unchecked);
    }

    // Insert all missing tables
    m_tableSelect->clear();
    for (const int &table : m_db->missingTables(round)) {
        m_tableSelect->addItem(QString::number(table), table);
    }

    // Insert all missing players/pairs

    for (int i = 1; i <= 2; i++) {
        m_playersSelect[i][1]->clear();
        if (m_db->getTournamentMode() == Database::TournamentMode::singlePlayers) {
            m_playersSelect[i][2]->clear();
        }
    }

    QMap<QString, int> missingPlayers = m_db->missingPlayers(round);
    QStringList missingPlayersNames = missingPlayers.keys();
    missingPlayersNames.sort(Qt::CaseInsensitive);
    for (const QString &missingPlayersName : missingPlayersNames) {
        for (int i = 1; i <= 2; i++) {
            m_playersSelect[i][1]->addItem(missingPlayersName,
                                           missingPlayers.value(missingPlayersName));
            if (m_db->getTournamentMode() == Database::TournamentMode::singlePlayers) {
                m_playersSelect[i][2]->addItem(missingPlayersName,
                                               missingPlayers.value(missingPlayersName));
            }
        }
    }

    if (m_roundsState.value(round)) {
        m_roundSummaryBox->setTitle(
            tr("Alle Ergebnisse aus Runde %1").arg(QString::number(round)));
    } else {
        m_roundSummaryBox->setTitle(
            tr("Bisher eingetragene Ergebnisse aus Runde %1").arg(QString::number(round)));
    }
    updateScore();

    m_assignment = m_db->getAssignment(round);
    blockSelectSignals(false);
    setScrollAreaHeight();
    tableChanged(0);
    m_searchWidget->restoreSearch();
}

void ScorePage::updateScore()
{
    const QList<Database::Score> score = m_db->getScores(
        m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt());
    m_scoreList->setRowCount(score.size() / m_db->getBoogersPerRound());

    if (score.size() == 0) {
        m_searchWidget->clear();
        m_isSearchLabel->hide();
        m_searchWidget->hide();
    } else {
        m_searchWidget->show();
    }

    int color = 1;
    int row = -1;
    for (int i = 0; i < score.size(); i++) {
        row++;
        color++;
        color = color % 2;
        m_scoreList->setItem(row, 0, new Cell(QString::number(score.at(i).table), false, color));
        m_scoreList->setSpan(row, 0, 2, 1);
        m_scoreList->setItem(row, 1, new Cell(score.at(i).pairName, false, color));
        m_scoreList->setItem(row, 2, new Cell(m_db->boogify(score.at(i).score), false, color));
        for (int booger = 2; booger <= m_db->getBoogersPerRound(); booger++) {
            m_scoreList->setItem(row, 1 + booger, new Cell(m_db->boogify(score.at(++i).score),
                                                         false, color));
        }
        row++;
        m_scoreList->setItem(row, 1, new Cell(score.at(++i).pairName, false, color));
        m_scoreList->setItem(row, 2, new Cell(m_db->boogify(score.at(i).score), false, color));
        for (int booger = 2; booger <= m_db->getBoogersPerRound(); booger++) {
            m_scoreList->setItem(row, 1 + booger, new Cell(m_db->boogify(score.at(++i).score),
                                                         false, color));
        }
    }
    m_scoreList->resizeColumnsToContents();
}

void ScorePage::tableChanged(int tableNumber)
{
    if (! m_autoSelect->isChecked()) {
        return;
    }

    int table = m_tableSelect->itemData(tableNumber).toInt();
    if (m_assignment.assignedPair.contains(table)) {
        if (m_assignment.assignedPair[table].contains(1)) {
            int target = m_playersSelect[1][1]->findData(m_assignment.assignedPair[table][1]);
            if (target != -1) {
                m_playersSelect[1][1]->setCurrentIndex(target);
            }
        }
        if (m_assignment.assignedPair[table].contains(2)) {
            int target = m_playersSelect[2][1]->findData(m_assignment.assignedPair[table][2]);
            if (target != -1) {
                m_playersSelect[2][1]->setCurrentIndex(target);
            }
        }
    }
}

void ScorePage::pair1Changed(int pair)
{
    pairChanged(1, pair);
}

void ScorePage::pair2Changed(int pair)
{
    pairChanged(2, pair);
}

void ScorePage::pair34Changed(int pair)
{
    pairChanged(0, pair);
}

void ScorePage::pairChanged(int pairNumber, int pair)
{
    if (m_autoSelect->isChecked()) {
        int pairId = m_playersSelect[pairNumber][1]->itemData(pair).toInt();
        if (m_assignment.assignedPairNumber.contains(pairId)) {
            if (pairNumber != m_assignment.assignedPairNumber.value(pairId)) {
                QMessageBox::warning(this,
                    tr("Die Paarauswahl passt nicht zur Auslosung"),
                    tr("<p>Als Paar %1 wurde gerade das Paar „%2“ ausgewählt.</p>"
                       "<p>Laut der Auslosung (bzw. dem Ergebnis der letzten Runde) "
                       "is dieses Paar aber ein Paar %3!</p>").arg(
                       QString::number(pairNumber),
                       m_playersSelect[pairNumber][1]->currentText(),
                       QString::number(m_assignment.assignedPairNumber.value(pairId))));
            } else {
                if (m_assignment.assignedTable.contains(pairId)) {
                    int target = m_tableSelect->findData(m_assignment.assignedTable.value(pairId));
                    if (target != -1) {
                        m_tableSelect->setCurrentIndex(target);
                    }
                }
            }
        }
    }

    bool savePossible = false;

    if (m_db->getTournamentMode() == Database::TournamentMode::fixedPairs) {
        if (m_playersSelect[1][1]->currentIndex()
            != m_playersSelect[2][1]->currentIndex()) {
            savePossible = true;
        }
    } else {
        QList<int> selectedPlayers;
        for (int i = 1; i <= 2; i++) {
            for (int j = 1; j <= 2; j++) {
                int selectedIndex = m_playersSelect[i][j]->currentIndex();
                if (! selectedPlayers.contains(selectedIndex)) {
                    selectedPlayers << selectedIndex;
                }
            }
        }
        if (selectedPlayers.size() == 4) {
            savePossible = true;
        }
    }

    m_saveScoreButton->setEnabled(savePossible);
}

void ScorePage::saveScore()
{
    emit statusUpdate(tr("Speichere Spielergebnis …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    bool updatePlayersPage = (! m_db->tournamentStarted());

    QList<int> players;
    QList<int> pair1Scores;
    QList<int> pair2Scores;

    for (int i = 1; i <= 2; i++) {
        players << m_playersSelect[i][1]->itemData(m_playersSelect[i][1]->currentIndex()).toInt();
    }
    if (m_db->getTournamentMode() == Database::TournamentMode::singlePlayers) {
        for (int i = 1; i <= 2; i++) {
            players << m_playersSelect[i][2]->itemData(
                           m_playersSelect[i][2]->currentIndex()).toInt();
        }
    }

    for (BoogerWidget *boogerWidget : m_boogerWidgets) {
        pair1Scores << boogerWidget->score(1);
        pair2Scores << boogerWidget->score(2);
    }

    int round = m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt();
    int table = m_tableSelect->itemData(m_tableSelect->currentIndex()).toInt();
    m_db->saveScore(round, table, players, pair1Scores, pair2Scores);

    if (updatePlayersPage) {
        emit tournamentStarted();
    }

    emit scoreChanged();
    m_saveScoreButton->setText(tr("Ergebnis\neintragen"));
    reload(round, m_abortRound->checkState());

    // Make sure the entered score is visible
    int targetRow = 0;
    for (int i = 0; i < m_scoreList->rowCount(); i += 2) {
        if (m_scoreList->item(i, 0)->text().toInt() == table) {
            targetRow = i;
            break;
        }
    }
    if (targetRow == m_scoreList->rowCount() - 2) {
        m_scoreList->scrollToBottom();
    } else {
        m_scoreList->scrollToItem(m_scoreList->item(targetRow, 1),
                                  QAbstractItemView::PositionAtCenter);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Spielergebnis für Tisch %1 in Runde %2 gespeichert").arg(
                         QString::number(table), QString::number(round)));
}

void ScorePage::triggerShowContextMenu()
{
    QPoint point = m_scoreList->mapFromGlobal(QCursor::pos());
    point.setY(point.y() - m_scoreList->horizontalHeader()->height());
    showContextMenu(point);
}

void ScorePage::showContextMenu(QPoint point)
{
    QTableWidgetItem *selectedScore = m_scoreList->currentItem();
    if (selectedScore == nullptr) {
        return;
    }

    int row = selectedScore->row();
    if (row % 2 != 0) {
        row = row - 1;
    }

    QString table = m_scoreList->item(row, 0)->text();
    m_menuTitle->setText(tr("Ergebnis für Tisch %1").arg(table));
    m_selectedTable = table.toInt();

    QPoint targetPoint = point;
    targetPoint.setY(targetPoint.y() + m_scoreList->horizontalHeader()->height());
    m_contextMenu->exec(m_scoreList->mapToGlobal(targetPoint));
}

void ScorePage::deleteScore()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Spielergebnis löschen"));
    messageBox.setText(tr("Soll das Spielergebnis von Tisch %1 wirklich gelöscht werden?").arg(
                          QString::number(m_selectedTable)));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    doDeleteScore(m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt(), m_selectedTable);
}

void ScorePage::doDeleteScore(int round, int table)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche Spielergebnis …"));

    m_db->deleteScore(round, table);
    reload();

    if (! m_db->tournamentStarted()) {
        // Re-emitting this will cause the players page to be updated again,
        // as it checks for db.tournamentStarted() itself
        emit tournamentStarted();
    }

    emit scoreChanged();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Spielergebnis für Tisch %1 in Runde %2 gelöscht").arg(
                         QString::number(round), QString::number(table)));
}

void ScorePage::editScore()
{
    int round = m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt();
    QList<Database::SingleScore> score = m_db->getSingleScore(round, m_selectedTable);
    doDeleteScore(round, m_selectedTable);
    m_roundSelect->setCurrentIndex(m_roundSelect->findData(round));

    blockSelectSignals(true);
    m_tableSelect->setCurrentIndex(m_tableSelect->findData(m_selectedTable));
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        setPairOrPlayer(1, 1, score[0].pairOrPlayer);
        setPairOrPlayer(2, 1, score[m_db->getBoogersPerRound()].pairOrPlayer);
        break;
    case Database::TournamentMode::singlePlayers:
        setPairOrPlayer(1, 1, score[0].pairOrPlayer);
        setPairOrPlayer(1, 2, score[m_db->getBoogersPerRound()].pairOrPlayer);
        setPairOrPlayer(2, 1, score[m_db->getBoogersPerRound() * 2].pairOrPlayer);
        setPairOrPlayer(2, 2, score[m_db->getBoogersPerRound() * 3].pairOrPlayer);
        break;
    }
    blockSelectSignals(false);

    QList<int> boogerScores = getBoogerScores(score);
    for (int i = 0; i < m_boogerWidgets.length(); i++) {
        m_boogerWidgets[i]->setScore(boogerScores[i], boogerScores[i + boogerScores.length() / 2]);
    }

    m_saveScoreButton->setEnabled(true);
    m_saveScoreButton->setFocus();
}

QList<int> ScorePage::getBoogerScores(QList<Database::SingleScore> &score)
{
    QList<int> boogerScores;
    QListIterator<Database::SingleScore> singleScore(score);
    while (singleScore.hasNext()) {
        boogerScores << singleScore.next().score;
        for (int i = 1; i < m_db->getBoogersPerRound(); i++) {
            boogerScores << singleScore.next().score;
        }
        if (m_db->getTournamentMode() == Database::TournamentMode::singlePlayers) {
            for (int i = 0; i < m_db->getBoogersPerRound(); i++) {
                singleScore.next();
            }
        }
    }
    return boogerScores;
}

void ScorePage::setPairOrPlayer(int pair, int player, int pairOrPlayer)
{
    m_playersSelect[pair][player]->setCurrentIndex(
        m_playersSelect[pair][player]->findData(pairOrPlayer)
    );
}

void ScorePage::blockSelectSignals(bool state)
{
    m_tableSelect->blockSignals(state);
    for (int i = 1; i <= 2; i++) {
        for (int j = 1; j <= 2; j++) {
            m_playersSelect[i][j]->blockSignals(state);
        }
    }
}

void ScorePage::checkEnterScore()
{
    if (! m_saveScoreButton->isEnabled()) {
        return;
    }
    saveScore();
}

void ScorePage::showEvent(QShowEvent *event)
{
    QWidget::showEvent(event);
    setScrollAreaHeight();
}

void ScorePage::updateSearch(QString searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_scoreList->rowCount(); row++) {
            if (m_scoreList->isRowHidden(row)) {
                m_scoreList->showRow(row);
            }
        }
        m_scoreList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    for (int row = 0; row < m_scoreList->rowCount(); row += 2) {
        if (m_scoreList->item(row, 1)->text().contains(searchText, Qt::CaseInsensitive)
            | m_scoreList->item(row + 1, 1)->text().contains(searchText, Qt::CaseInsensitive)) {
            if (m_scoreList->isRowHidden(row)) {
                m_scoreList->showRow(row);
                m_scoreList->showRow(row + 1);
            }
            matches++;
        } else {
            if (! m_scoreList->isRowHidden(row)) {
                m_scoreList->hideRow(row);
                m_scoreList->hideRow(row + 1);
            }
        }
    }

    m_searchWidget->updateMatches(matches);
    m_scoreList->resizeColumnsToContents();

    if (! m_isSearchLabel->isVisible()) {
        m_isSearchLabel->show();
    }
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
}

void ScorePage::swapPairs()
{
    int round = m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt();
    QList<Database::SingleScore> score = m_db->getSingleScore(round, m_selectedTable);
    m_db->deleteScore(round, m_selectedTable);

    QList<int> boogerScores = getBoogerScores(score);
    QList<int> players;
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        players << score[m_db->getBoogersPerRound()].pairOrPlayer
                << score[0].pairOrPlayer;
        break;
    case Database::TournamentMode::singlePlayers:
        players << score[m_db->getBoogersPerRound() * 2].pairOrPlayer
                << score[0].pairOrPlayer
                << score[m_db->getBoogersPerRound() * 3].pairOrPlayer
                << score[m_db->getBoogersPerRound()].pairOrPlayer;
        break;
    }
    m_db->saveScore(round, m_selectedTable, players,
                    boogerScores.mid(boogerScores.length() / 2, boogerScores.length() / 2),
                    boogerScores.mid(0, boogerScores.length() / 2));

    emit statusUpdate(tr("Paar 1 und Paar 2 für Tisch %1 in Runde %2 vertauscht").arg(
                         QString::number(round), QString::number(m_selectedTable)));
    reload();
}
