/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "TournamentInfoDialog.h"
#include "Database.h"

// Qt includes
#include <QGridLayout>
#include <QLabel>
#include <QDir>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QDialogButtonBox>
#include <QPushButton>
#include <QDebug>

TournamentInfoDialog::TournamentInfoDialog(QWidget *parent, Database *db)
    : QDialog(parent), m_db(db)
{
    setAttribute(Qt::WA_DeleteOnClose);
    setWindowTitle(tr("Muckturnier: Turnierinformationen"));

    m_mainLayout = new QVBoxLayout;
    setLayout(m_mainLayout);

    addGroup(tr("Datenbank"));

    addLabel(tr("Dateiname:"));
    QString dbFile = QDir::toNativeSeparators(m_db->getDbFile());
    QFont font = this->font();
    QFontMetrics fontMetrics(font);
    if (fontMetrics.boundingRect(dbFile).width() > fontMetrics.boundingRect(
        QString::fromUtf8("/A/very/long/path/which/still/does/not/mess/up/the/layout")).width()) {
        addInfo(dbFile, true);
    } else {
        addInfo(dbFile, false);
    }

    addLabel(tr("Erstellt mit:"));
    addInfo(tr("Muckturnier %1").arg(m_db->getMuckturnierVersion()));

    addGroup(tr("Einstellungen"));

    addLabel(tr("Tuniermodus:"));
    switch (m_db->getTournamentMode()) {
    case Database::fixedPairs:
        addInfo(tr("Feste Paare"));
        break;
    case Database::singlePlayers:
        addInfo(tr("Einzelne Spieler"));
        break;
    }

    addLabel(tr("Punkte pro Bobbl:"));
    addInfo(QString::number(m_db->getBoogerScore()));

    addLabel(tr("Bobbl pro Runde:"));
    addInfo(QString::number(m_db->getBoogersPerRound()));

    addGroup(tr("Turnier"));

    addLabel(tr("Angemeldete %1:").arg(m_db->playersString(Database::Numerus::plural)));
    addInfo(QString::number(m_db->playersCount()));

    addLabel(tr("Gespielte Runden:"));
    QMap<int, bool> roundsState = m_db->roundsState();
    int rounds = roundsState.count();
    if (rounds == 1) {
        QList<Database::Score> scores = m_db->getScores(1);
        if (scores.count() == 0) {
            rounds = 0;
        }
    }
    addInfo(QString::number(rounds));

    bool allFinished = true;
    for (const bool &finished : roundsState.values()) {
        if (! finished) {
            allFinished = false;
            break;
        }
    }
    addLabel(tr("Alle Runden abgeschlossen:"));
    if (allFinished) {
        addInfo(tr("Ja"));
    } else {
        addInfo(tr("Nein"));
    }

    QDialogButtonBox *buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok);
    connect(buttonBox->button(QDialogButtonBox::Ok), &QPushButton::clicked,
            this, &QDialog::accept);
    m_mainLayout->addWidget(buttonBox);
}

void TournamentInfoDialog::addGroup(QString text)
{
    QGroupBox *group = new QGroupBox(text);
    m_mainLayout->addWidget(group);
    QGridLayout *layout = new QGridLayout(group);
    m_groupLayout = layout;
    m_row = -1;
}

void TournamentInfoDialog::addLabel(QString text)
{
    m_row++;
    QLabel *label = new QLabel(text);
    label->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Minimum);
    label->setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByMouse);
    label->setAlignment(Qt::AlignTop);
    m_groupLayout->addWidget(label, m_row, 0);
}

void TournamentInfoDialog::addInfo(QString text, bool wordWrap)
{
    QLabel *label = new QLabel(text);
    label->setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByMouse);
    if (wordWrap) {
        label->setWordWrap(true);
    }
    m_groupLayout->addWidget(label, m_row, 1);
}
